# Copyright (c) 2019 Eric Steinberger


import copy

import numpy as np

from PokerRL.game.Poker import Poker
from PokerRL.game._.wrappers._Wrapper import Wrapper


class FlatHULimitPokerHistoryWrapper(Wrapper):
    """
    This wrapper only supports Heads Up games.
    This wrapper is suitable for feedforward NN architectures.

    Stores action sequence and appends to current obs similarly as is proposed in https://arxiv.org/abs/1603.01121
    Action History is *AFTER* raw env obs state in the vector
    """

    def __init__(self, env, env_bldr_that_built_me):
        assert env.N_SEATS == 2
        super().__init__(env=env, env_bldr_that_built_me=env_bldr_that_built_me)

        self._action_vector_size = env_bldr_that_built_me.action_vector_size
        self._action_count_this_round = None
        self._game_round_last_tick = None
        self._action_history_vector = None

    def _reset_state(self, **kwargs):
        self._action_count_this_round = [0, 0]  # one per player
        self._game_round_last_tick = Poker.PREFLOP
        self._action_history_vector = np.zeros(shape=self._action_vector_size, dtype=np.float32)

    def _pushback(self, env_obs=None):
        # In case of a new round being dealt, the last action still has to have been in the old round since we observe
        # every transition. That means that all the new_round logic has to be executed from the *next* transition
        # onwards; this transition is handled within the old round.

        # If None, env was just reset
        _last_a = self.env.last_action[0]
        if _last_a is not None:
            _last_actor = self.env.last_action[2]

            idx = self.env_bldr.get_vector_idx(round_=self._game_round_last_tick,
                                               p_id=_last_actor,
                                               nth_action_this_round=self._action_count_this_round[_last_actor],
                                               action_idx=_last_a)

            self._action_history_vector[idx] = 1  # one-hot over actions

            self._action_count_this_round[_last_actor] += 1
            if self.env.current_round != self._game_round_last_tick:
                self._game_round_last_tick = self.env.current_round
                self._action_count_this_round = [0, 0]

    def print_obs(self, wrapped_obs=None):
        if wrapped_obs is None:
            wrapped_obs = self.get_current_obs()

        assert isinstance(wrapped_obs, np.ndarray)
        print()
        print()
        print("*****************************************************************************************************")
        print()
        print("________________________________________ OBSERVATION HISTORY ________________________________________")
        print()
        self.env.print_obs(wrapped_obs[:self.env_bldr.pub_obs_size - self._action_vector_size])
        print()
        print("------------------- Action Sequence --------------------")
        print(wrapped_obs[self.env_bldr.pub_obs_size - self._action_vector_size:])

    def get_current_obs(self, env_obs=None):
        if env_obs is None:
            return np.concatenate((self.env.get_current_obs(is_terminal=False), self._action_history_vector,), axis=0)
        else:
            return np.concatenate((env_obs, self._action_history_vector,), axis=0)

    def state_dict(self):
        return {
            "base": super().state_dict(),
            "a_seq": np.copy(self._action_history_vector),
            "game_round_last_tick": self._game_round_last_tick,
            "action_count_this_round": copy.deepcopy(self._action_count_this_round),
        }

    def load_state_dict(self, state_dict):
        super().load_state_dict(state_dict=state_dict["base"])
        self._action_history_vector = np.copy(state_dict["a_seq"])
        self._game_round_last_tick = state_dict["game_round_last_tick"]
        self._action_count_this_round = copy.copy(state_dict["action_count_this_round"])

    def set_to_public_tree_node_state(self, node):
        """
        Sets the internal env wrapper to the state ""node"" is in.

        Args:
            node:                   Any node (of any type) in a PublicTree instance.
        """
        state_seq = []  # will be sorted. [0] is root.

        def add(_node):
            if _node is not None:
                if _node.p_id_acting_next != _node.tree.CHANCE_ID:
                    state_seq.insert(0, _node.env_state)
                add(_node.parent)

        add(node)  # will go from node to parent to parent of parent... to root and reverse the direction for the tree
        # fake step in the internal env to gain current rr state
        self.reset()
        self._reset_state()
        for sd in state_seq:
            self.env.load_state_dict(sd, blank_private_info=True)  # load cause _pushback uses the env's internal state
            self._pushback()



# Copyright (c) 2019 Eric Steinberger


import numpy as np

from PokerRL.game.Poker import Poker
from PokerRL.game._.rl_env.base.PokerEnv import PokerEnv as _PokerEnv
from PokerRL.game.poker_env_args import LimitPokerEnvArgs


class LimitPokerEnv(_PokerEnv):
    """
    Adapts the dynamics of PokerEnv to fit to fixed-limit rules.
    """

    ARGS_CLS = LimitPokerEnvArgs

    def __init__(self,
                 env_args,
                 is_evaluating,
                 lut_holder):
        assert isinstance(env_args, LimitPokerEnvArgs)
        super().__init__(env_args=env_args, lut_holder=lut_holder, is_evaluating=is_evaluating)
        self.uniform_action_interpolation = False
        self.N_ACTIONS = env_args.N_ACTIONS

    def _get_env_adjusted_action_formulation(self, action_int):
        if action_int == 0:
            return [0, -1]
        if action_int == 1:
            return [1, -1]
        elif action_int == 2:
            return [2, -1]  # has to be fixed in a call of ._get_fixed_a() !!!
        else:
            raise ValueError(action_int)

    def _adjust_raise(self, raise_total_amount_in_chips):
        b = self.BIG_BET if self.current_round >= self.ROUND_WHERE_BIG_BET_STARTS else self.SMALL_BET
        return (self.n_raises_this_round + 1) * b

    def get_legal_actions(self):
        """

        Returns:
            list:   a list with none, one or multiple actions of the set [FOLD, CHECK/CALL, BET]
        """
        legal_actions = []
        for a_int in [Poker.FOLD, Poker.CHECK_CALL]:
            _a = self._get_env_adjusted_action_formulation(action_int=a_int)
            if self._get_fixed_action(action=_a)[0] == a_int:
                legal_actions.append(a_int)

        adj_a = self._get_env_adjusted_action_formulation(action_int=Poker.BET_RAISE)
        fixed_a = self._get_fixed_action(action=adj_a)
        if ((self.n_raises_this_round < self.MAX_N_RAISES_PER_ROUND[self.current_round])
            and (adj_a[0] == fixed_a[0])):
            legal_actions.append(Poker.BET_RAISE)

        return legal_actions

    def get_hand_rank(self, hand_2d, board_2d):
        raise NotImplementedError

    def get_hand_rank_all_hands_on_given_boards(self, boards_1d, lut_holder):
        """
        This feature can allows batch computing.
        """
        raise NotImplementedError

    def get_random_action(self):
        legal = self.get_legal_actions()
        return legal[np.random.randint(len(legal))]

    def print_tutorial(self):
        print("____________________________________________ TUTORIAL ____________________________________________")
        print("Actions:")
        print("0 \tFold")
        print("1 \tCall")
        for i in range(2, self.N_ACTIONS):
            print(i, "\tRaise according to current fixed limit")

    def human_api_ask_action(self):
        """ Returns action in Tuple form. """
        while True:
            try:
                action_idx = int(
                    input("What action do you want to take as player " + str(self.current_player.seat_id) + "?"))
            except ValueError:
                print("You need to enter one of the allowed actions. Refer to the tutorial please.")
                continue
            if action_idx not in [Poker.FOLD, Poker.CHECK_CALL, Poker.BET_RAISE]:
                print("Invalid action_idx! Please enter one of the allowed actions as described in the tutorial above.")
                continue
            break
        return self._get_fixed_action(self._get_env_adjusted_action_formulation(action_int=action_idx))



# Copyright (c) 2019 Eric Steinberger


from PokerRL.rl import rl_util
from PokerRL.rl.MaybeRay import MaybeRay
from PokerRL.util.file_util import do_pickle, load_pickle


class EvalAgentBase:
    """
    This baseclass should be subclassed by each agent/algorithm type. It is used to wrap the agent with his own
    internal1 environment. It provides a standardized API for querying the agent for different things to the evaluators.
    If an algorithm employs different agents for each seat on the table, this class should wrap all of them in one.
    """
    ALL_MODES = NotImplementedError  # Override with list of all modes

    def __init__(self, t_prof, mode=None, device=None):
        """
        Args:
            t_prof (TrainingProfile):
            mode:                       Any mode your algorithm's eval agent can be evaluated in. Specify modes
                                        as class variables and pass one of them here. Can be changed later by calling
                                        .to_mode(new_mode) on this instance
            device (torch.device):      The device the eval agent shall live and act on.
        """
        self.t_prof = t_prof
        self.ray = MaybeRay(runs_distributed=t_prof.DISTRIBUTED, runs_cluster=t_prof.CLUSTER)
        self.env_bldr = rl_util.get_env_builder(t_prof=t_prof)

        self._internal_env_wrapper = self.env_bldr.get_new_wrapper(is_evaluating=True, stack_size=None)
        self._mode = mode

        if device is None:
            self.device = self.t_prof.device_inference
        else:
            self.device = device

    # __________________________________________________ Query agents __________________________________________________
    def get_a_probs_for_each_hand(self):
        """
        Returns:
            np.ndarray(RANGE_SIZE, N_ACTIONS): the action probabilities for each hand
        """
        raise NotImplementedError

    def get_a_probs(self):
        """
        Returns:
            np.ndarray(N_ACTIONS): action probs for hand currently held in current state
        """
        raise NotImplementedError

    def get_action(self, step_env=True, need_probs=False):
        """
        Args:
            step_env (bool):        Whether the internal env shall be stepped
            need_probs (bool):      Whether the action probabilities for all hands shall be returned too

        Returns:
            action,
            action probs for each hand (or None if not need_probs)
        """
        raise NotImplementedError

    def state_dict(self):
        """ Override and keep base as one field! """
        return {
            "t_prof": self.t_prof,
            "mode": self._mode,
            "env": self._internal_env_wrapper.state_dict(),
            "agent": self._state_dict(),
        }

    def load_state_dict(self, state):
        self._internal_env_wrapper.load_state_dict(state["env"])
        self._mode = state["mode"]
        self._load_state_dict(state["agent"])

    def _state_dict(self):
        # Implement your agent's state_dict
        raise NotImplementedError

    def _load_state_dict(self, state):
        # Implement your agent's load_state_dict
        raise NotImplementedError

    def update_weights(self, weights_for_eval_agent):
        """
        Args:
            weights_for_eval_agent: Can be any algorithm-specific data; e.g. Neural Network parameters for the agent
        """
        raise NotImplementedError

    def can_compute_mode(self):
        """
        Returns:
            bool:                   Whether whatever condition is satisfied (e.g. for delayed CFR+ whether enough
                                    iterations have passed) to evaluate the algorithm with self._mode
        """
        raise NotImplementedError

    # _____________________________________________________ State ______________________________________________________
    def set_stack_size(self, stack_size):
        self._internal_env_wrapper.env.set_stack_size(stack_size=stack_size)

    def get_mode(self):
        return self._mode

    def set_mode(self, mode):
        assert mode in self.ALL_MODES
        self._mode = mode

    def set_env_wrapper(self, env_wrapper):
        self._internal_env_wrapper = env_wrapper

    def get_env_wrapper(self):
        return self._internal_env_wrapper

    def set_to_public_tree_node_state(self, node):
        self._internal_env_wrapper.set_to_public_tree_node_state(node=node)

    # __________________________________________________ Notifications _________________________________________________
    def notify_of_action(self, p_id_acted, action_he_did):
        assert self._internal_env_wrapper.env.current_player.seat_id == p_id_acted
        self._internal_env_wrapper.step(action=action_he_did)

    def notify_of_processed_tuple_action(self, p_id_acted, action_he_did):
        assert self._internal_env_wrapper.env.current_player.seat_id == p_id_acted
        self._internal_env_wrapper.step_from_processed_tuple(action=action_he_did)

    def notify_of_raise_frac_action(self, p_id_acted, frac):
        """ this fn is only useful to call if current_player wants to raise. Therefore it assumes that's the case. """
        assert self._internal_env_wrapper.env.current_player.seat_id == p_id_acted
        self._internal_env_wrapper.step_raise_pot_frac(pot_frac=frac)

    def notify_of_reset(self):
        self._internal_env_wrapper.reset()
        self._internal_env_wrapper._list_of_obs_this_episode = []  # from .reset() the first obs is in by default

    def reset(self, deck_state_dict=None):
        self._internal_env_wrapper.reset(deck_state_dict=deck_state_dict)

    # ___________________________________________________ Store State __________________________________________________
    def env_state_dict(self):
        return self._internal_env_wrapper.state_dict()

    def load_env_state_dict(self, state_dict):
        self._internal_env_wrapper.load_state_dict(state_dict)

    def store_to_disk(self, path, file_name):
        do_pickle(obj=self.state_dict(), path=path, file_name=file_name)

    @classmethod
    def load_from_disk(cls, path_to_eval_agent):
        state = load_pickle(path=path_to_eval_agent)

        eval_agent = cls(t_prof=state["t_prof"])
        eval_agent.load_state_dict(state=state)

        return eval_agent



# Copyright (c) 2019 Eric Steinberger


from PokerRL.rl.rl_util import get_builder_from_str, get_env_cls_from_str


def get_env_builder_lbr(t_prof):
    env_bldr_cls = get_builder_from_str(t_prof.env_builder_cls_str)
    return env_bldr_cls(env_cls=get_env_cls_from_str(t_prof.game_cls_str),
                        env_args=t_prof.module_args["lbr"].get_lbr_env_args(agents_env_args=t_prof.module_args["env"]))



# Copyright (c) 2019 Eric Steinberger


import unittest
from unittest import TestCase

import numpy as np

from PokerRL.game.Poker import Poker
from PokerRL.game._.look_up_table import LutHolderHoldem, _LutGetterHoldem, _LutGetterLeduc
from PokerRL.game.games import StandardLeduc, DiscretizedNLHoldem


class TestLutGetterHoldem(TestCase):

    def test_get_1d_card_2_2d_card_lut(self):
        lg = _LutGetterHoldem(env_cls=DiscretizedNLHoldem)
        lut = lg.get_1d_card_2_2d_card_LUT()
        assert lut.shape == (52, 2)
        assert not np.any(lut == -2)

    def test_get_2d_card_2_1d_card_lut(self):
        lg = _LutGetterHoldem(env_cls=DiscretizedNLHoldem)
        lut = lg.get_2d_card_2_1d_card_LUT()
        assert lut.shape == (13, 4)
        assert not np.any(lut == -2)

    def test_get_idx_2_hole_card_lut(self):
        lg = _LutGetterHoldem(env_cls=DiscretizedNLHoldem)
        lut = lg.get_idx_2_hole_card_LUT()
        assert lut.shape == (1326, 2)
        assert not np.any(lut == -2)

    def test_get_hole_card_2_idx_lut(self):
        lg = _LutGetterHoldem(env_cls=DiscretizedNLHoldem)
        lut = lg.get_hole_card_2_idx_LUT()
        assert lut.shape == (52, 52)
        for i in range(52):
            for i2 in range(i + 1, 52):
                assert lut[i, i2] != -2
            for _i2 in range(0, i):
                assert lut[i, _i2] == -2

    def test_get_lut_card_in_what_range_idxs(self):
        lg = _LutGetterHoldem(env_cls=DiscretizedNLHoldem)
        lut = lg.get_card_in_what_range_idxs_LUT()
        assert lut.shape == (52, 51)
        assert not np.any(lut == -2)

        counts = np.zeros(1326, np.int32)

        for c in range(52):
            for h in lut[c]:
                counts[h] += 1

        assert np.all(counts == 2)


class TestLutGetterLeduc(TestCase):

    def test_get_1d_card_2_2d_card_lut(self):
        lg = _LutGetterLeduc(env_cls=StandardLeduc)
        lut = lg.get_1d_card_2_2d_card_LUT()
        assert lut.shape == (6, 2)
        assert not np.any(lut == -2)

    def test_get_2d_card_2_1d_card_lut(self):
        lg = _LutGetterLeduc(env_cls=StandardLeduc)
        lut = lg.get_2d_card_2_1d_card_LUT()
        assert lut.shape == (3, 2)
        assert not np.any(lut == -2)

    def test_get_idx_2_hole_card_lut(self):
        lg = _LutGetterLeduc(env_cls=StandardLeduc)
        lut = lg.get_idx_2_hole_card_LUT()
        assert lut.shape == (6, 1)
        assert not np.any(lut == -2)

    def test_get_hole_card_2_idx_lut(self):
        lg = _LutGetterLeduc(env_cls=StandardLeduc)
        lut = lg.get_hole_card_2_idx_LUT()
        assert lut.shape == (6, 1)
        assert not np.any(lut == -2)

    def test_get_lut_card_in_what_range_idxs(self):
        lg = _LutGetterLeduc(env_cls=StandardLeduc)
        lut = lg.get_card_in_what_range_idxs_LUT()
        assert lut.shape == (6, 1)
        assert not np.any(lut == -2)

        counts = np.zeros(6, np.int32)

        for c in range(6):
            for h in lut[c]:
                counts[h] += 1

        assert np.all(counts == 1)


class TestLutHolderHoldem(TestCase):

    def test_create(self):
        lh = DiscretizedNLHoldem.get_lut_holder()

        assert lh.LUT_1DCARD_2_2DCARD.dtype == np.dtype(np.int8)
        assert lh.LUT_2DCARD_2_1DCARD.dtype == np.dtype(np.int8)
        assert lh.LUT_IDX_2_HOLE_CARDS.dtype == np.dtype(np.int8)
        assert lh.LUT_HOLE_CARDS_2_IDX.dtype == np.dtype(np.int16)

    def test_get_1d_card(self):
        lh = DiscretizedNLHoldem.get_lut_holder()
        assert lh.get_1d_card(card_2d=[0, 3]) == 3
        assert lh.get_1d_card(card_2d=[12, 3]) == 51
        assert lh.get_1d_card(card_2d=np.array([0, 0], dtype=np.int8)) == 0
        assert lh.get_1d_card(card_2d=np.array([0, 0], dtype=np.int32)) == 0
        assert lh.get_1d_card(card_2d=Poker.CARD_NOT_DEALT_TOKEN_2D) == Poker.CARD_NOT_DEALT_TOKEN_1D

    def test_get_1d_cards(self):
        lh = DiscretizedNLHoldem.get_lut_holder()
        assert np.array_equal(lh.get_1d_cards(cards_2d=np.array([[0, 3]])), np.array([3]))
        assert np.array_equal(lh.get_1d_cards(cards_2d=np.array([[0, 3], [12, 3]])), np.array([3, 51]))
        assert np.array_equal(lh.get_1d_cards(cards_2d=np.array([])), np.array([], np.int8))
        assert np.array_equal(lh.get_1d_cards(
            cards_2d=np.concatenate((np.array([[0, 0]]), Poker.CARD_NOT_DEALT_TOKEN_2D.reshape(-1, 2)))),
            np.array([0, Poker.CARD_NOT_DEALT_TOKEN_1D], dtype=np.int8))

    def test_get_2d_cards(self):
        lh = DiscretizedNLHoldem.get_lut_holder()
        assert np.array_equal(lh.get_2d_cards(cards_1d=np.array([])), np.array([]))
        assert np.array_equal(lh.get_2d_cards(cards_1d=np.array([3])), np.array([[0, 3]]))
        assert np.array_equal(lh.get_2d_cards(cards_1d=np.array([3, 51])), np.array([[0, 3], [12, 3]]))
        assert np.array_equal(lh.get_2d_cards(cards_1d=np.array([])), np.array([]))
        assert np.array_equal(lh.get_2d_cards(cards_1d=np.array([0, Poker.CARD_NOT_DEALT_TOKEN_1D], dtype=np.int8)),
                              np.concatenate((np.array([[0, 0]]), Poker.CARD_NOT_DEALT_TOKEN_2D.reshape(-1, 2))))

    def test_get_range_idx_from_hole_cards(self):
        lh = DiscretizedNLHoldem.get_lut_holder()
        n = 0
        for c1 in range(51):
            for c2 in range(c1 + 1, 52):
                assert lh.LUT_HOLE_CARDS_2_IDX[c1, c2] == n

                n += 1

    def test_hole_card_luts(self):
        """ tests reversibility """
        lh = DiscretizedNLHoldem.get_lut_holder()
        for h in range(1326):
            _c_1d = lh.get_1d_hole_cards_from_range_idx(h)  # Tests 1d conversion
            _c_2d = lh.get_2d_hole_cards_from_range_idx(h)  # Tests 2d conversion

            c_1d = (lh.LUT_IDX_2_HOLE_CARDS[h])
            c_2d = np.array([lh.LUT_1DCARD_2_2DCARD[c_1d[0]],
                             lh.LUT_1DCARD_2_2DCARD[c_1d[1]]],
                            dtype=np.int8)

            assert np.array_equal(c_1d, _c_1d)
            assert np.array_equal(c_2d, _c_2d)

        # Tests inverse and thus validates that the mapping is unique for both 1d and 2d implicitly
        for c1 in range(51):
            for c2 in range(c1 + 1, 50):
                cc1 = lh.LUT_1DCARD_2_2DCARD[c1]
                cc2 = lh.LUT_1DCARD_2_2DCARD[c2]
                hole_cards = np.array([cc1, cc2], dtype=np.int8)
                assert lh.get_range_idx_from_hole_cards(hole_cards) == \
                       lh.LUT_HOLE_CARDS_2_IDX[c1, c2]


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Eric Steinberger


import copy

from PokerRL.game.poker_env_args import DiscretizedPokerEnvArgs, LimitPokerEnvArgs, NoLimitPokerEnvArgs
from PokerRL.rl.agent_modules.DDQN import DDQNArgs
from PokerRL.rl.neural.DuelingQNet import DuelingQArgs


class RLBRArgs:

    def __init__(self,
                 rlbr_bet_set,
                 n_hands_each_seat=20000,
                 n_workers=5,

                 # Training
                 DISTRIBUTED=False,
                 n_iterations=10000,
                 play_n_games_per_iter=50,
                 pretrain_n_games=5120,
                 device_training="cpu",

                 # the DDQN
                 nn_type="feedforward",
                 target_net_update_freq=300,
                 batch_size=512,
                 buffer_size=5e4,
                 optim_str="adam",
                 loss_str="mse",
                 lr=0.001,
                 eps_start=0.3,
                 eps_min=0.02,
                 eps_const=0.02,
                 eps_exponent=0.7,

                 # the QNet
                 rnn_cls_str="lstm",
                 n_units_final=64,
                 n_merge_and_table_layer_units=64,
                 n_cards_state_units=192,
                 use_pre_layers=False,
                 rnn_units=32,
                 rnn_stack=1,
                 dropout=0.0,
                 ):

        if nn_type == "recurrent":
            from PokerRL.rl.neural.MainPokerModuleRNN import MPMArgsRNN

            mpm_args = MPMArgsRNN(rnn_cls_str=rnn_cls_str,
                                  rnn_units=rnn_units,
                                  rnn_stack=rnn_stack,
                                  rnn_dropout=dropout,
                                  use_pre_layers=use_pre_layers,
                                  n_cards_state_units=n_cards_state_units,
                                  n_merge_and_table_layer_units=n_merge_and_table_layer_units)

        elif nn_type == "feedforward":
            from PokerRL.rl.neural.MainPokerModuleFLAT import MPMArgsFLAT

            mpm_args = MPMArgsFLAT(use_pre_layers=use_pre_layers,
                                   card_block_units=n_cards_state_units,
                                   other_units=n_merge_and_table_layer_units)

        else:
            raise ValueError(nn_type)
        if DISTRIBUTED and n_workers < 2:
            raise RuntimeError("RL-BR needs at least 2 workers, when running distributed. This is for 1 ParameterServer"
                               "and at least one LearnerActor")

        self.n_las = (n_workers - 1) if DISTRIBUTED else 1

        self.n_hands_each_seat = int(n_hands_each_seat)
        self.n_iterations = int(n_iterations)
        self.play_n_games_per_iter = int(play_n_games_per_iter)
        self.pretrain_n_games = int(pretrain_n_games)
        self.rlbr_bet_set = rlbr_bet_set

        self.ddqn_args = DDQNArgs(
            q_args=DuelingQArgs(
                n_units_final=n_units_final,
                mpm_args=mpm_args),
            cir_buf_size=int(buffer_size),
            batch_size=int(batch_size),
            n_mini_batches_per_update=1,
            target_net_update_freq=target_net_update_freq,
            optim_str=optim_str,
            loss_str=loss_str,
            lr=lr,
            eps_start=eps_start,
            eps_const=eps_const,
            eps_exponent=eps_exponent,
            eps_min=eps_min,
            grad_norm_clipping=1.0,
            device_training=device_training,
        )

    def get_rlbr_env_args(self, agents_env_args, randomization_range=None):
        arg_cls = type(agents_env_args)

        if arg_cls is DiscretizedPokerEnvArgs:
            return DiscretizedPokerEnvArgs(
                n_seats=agents_env_args.n_seats,
                starting_stack_sizes_list=copy.deepcopy(agents_env_args.starting_stack_sizes_list),
                bet_sizes_list_as_frac_of_pot=copy.deepcopy(self.rlbr_bet_set),
                stack_randomization_range=randomization_range if randomization_range else (0, 0),
                use_simplified_headsup_obs=agents_env_args.use_simplified_headsup_obs,
                uniform_action_interpolation=False
            )

        elif arg_cls is LimitPokerEnvArgs:
            return LimitPokerEnvArgs(
                n_seats=agents_env_args.n_seats,
                starting_stack_sizes_list=copy.deepcopy(agents_env_args.starting_stack_sizes_list),
                stack_randomization_range=randomization_range if randomization_range else (0, 0),
                use_simplified_headsup_obs=agents_env_args.use_simplified_headsup_obs,
                uniform_action_interpolation=False
            )

        elif arg_cls is NoLimitPokerEnvArgs:
            raise NotImplementedError("Currently not supported")

        else:
            raise NotImplementedError(arg_cls)



# Copyright (c) 2019 Eric Steinberger


import torch.nn as nn


class QNet(nn.Module):

    def __init__(self, env_bldr, q_args, device):
        super().__init__()

        self._env_bldr = env_bldr
        self._args = q_args
        self._n_actions = env_bldr.N_ACTIONS

        self._relu = nn.ReLU(inplace=False)

        MPM = q_args.mpm_args.get_mpm_cls()
        self._mpm = MPM(env_bldr=env_bldr, device=device, mpm_args=q_args.mpm_args)

        self._final_layer = nn.Linear(in_features=self._mpm.output_units,
                                      out_features=q_args.n_units_final)

        self._val = nn.Linear(in_features=q_args.n_units_final, out_features=self._n_actions)

        self.to(device)

    def forward(self, pub_obses, range_idxs, legal_action_masks):
        """
        Args:
            pub_obses (list):                       list of np arrays of shape [np.arr([history_len, n_features]), ...)
            range_idxs (torch.Tensor):              integer representation of hand
        """
        y = self._mpm(pub_obses=pub_obses, range_idxs=range_idxs)
        y = self._relu(self._final_layer(y))
        y = self._val(y)

        # The mask is important because the strategy inference assumes illegals to have exactly zero
        return y * legal_action_masks


class QNetArgs:

    def __init__(self, n_units_final, mpm_args):
        self.n_units_final = n_units_final
        self.mpm_args = mpm_args



# Copyright (c) 2019 Eric Steinberger


"""
This script runs 150 iterations of CFR in an abstracted version of Discrete No-Limit Leduc Poker with actions
{FOLD, CHECK/CALL, POT-SIZE-BET/RAISE}.
"""

from PokerRL.cfr.VanillaCFR import VanillaCFR
from PokerRL.game import bet_sets
from PokerRL.game.games import DiscretizedNLLeduc
from PokerRL.rl.base_cls.workers.ChiefBase import ChiefBase

if __name__ == '__main__':
    from PokerRL._.CrayonWrapper import CrayonWrapper

    n_iterations = 150
    name = "CFR_EXAMPLE"

    # Passing None for t_prof will is enough for ChiefBase. We only use it to log; This CFR impl is not distributed.
    chief = ChiefBase(t_prof=None)
    crayon = CrayonWrapper(name=name,
                           path_log_storage=None,
                           chief_handle=chief,
                           runs_distributed=False,
                           runs_cluster=False,
                           )
    cfr = VanillaCFR(name=name,
                     game_cls=DiscretizedNLLeduc,
                     agent_bet_set=bet_sets.POT_ONLY,
                     chief_handle=chief)

    for iter_id in range(n_iterations):
        print("Iteration: ", iter_id)
        cfr.iteration()
        crayon.update_from_log_buffer()
        crayon.export_all(iter_nr=iter_id)



# Copyright (c) 2019 Eric Steinberger


"""
This file is not runable; it's is a template to show how you could play against your algorithms. To do so,
replace "YourAlgorithmsEvalAgentCls" with the EvalAgent subclass (not instance) of your algorithm.

Note that you can see the AI's cards on the screen since this is just a research application and not meant for actual
competition. The AI can, of course, NOT see your cards.
"""

from PokerRL.game.InteractiveGame import InteractiveGame

if __name__ == '__main__':
    eval_agent = YourAlgorithmsEvalAgentCls.load_from_disk(
        path_to_eval_agent="\\path\\to\\your\\eval_agent\\eval_agent.pkl")

    game = InteractiveGame(env_cls=eval_agent.env_bldr.env_cls,
                           env_args=eval_agent.env_bldr.env_args,
                           seats_human_plays_list=[0],
                           eval_agent=eval_agent,
                           )

    game.start_to_play()



import numpy as np

from PokerRL.eval.rl_br import _util
from PokerRL.rl import rl_util
from PokerRL.rl.agent_modules.DDQN import DDQN
from PokerRL.rl.base_cls.workers.WorkerBase import WorkerBase


class Local_RLBR_LearnerActor(WorkerBase):

    def __init__(self, t_prof, chief_handle, eval_agent_cls):
        super().__init__(t_prof=t_prof)
        self._args = t_prof.module_args["rlbr"]

        self._env_bldr = rl_util.get_env_builder(t_prof=t_prof)

        self._chief_handle = chief_handle
        self._eval_agent_cls = eval_agent_cls
        self._eval_env_bldr = _util.get_env_builder_rlbr(t_prof=t_prof)

        self._ddqns = [None for _ in range(self._eval_env_bldr.N_SEATS)]
        self._rlbr_seat_id = None
        self._agent_seat_id = None
        self._rlbr_env_wrapper = None
        self._opponent = None
        self._buf = None
        self._br_memory_saver = None

        if t_prof.nn_type == "recurrent":
            from PokerRL.rl.buffers.CircularBufferRNN import CircularBufferRNN
            from PokerRL.rl.buffers.BRMemorySaverRNN import BRMemorySaverRNN

            self.CircularBufferCls = CircularBufferRNN
            self.BRMemorySaverCls = BRMemorySaverRNN
        elif t_prof.nn_type == "feedforward":
            from PokerRL.rl.buffers.CircularBufferFLAT import CircularBufferFLAT
            from PokerRL.rl.buffers.BRMemorySaverFLAT import BRMemorySaverFLAT

            self.CircularBufferCls = CircularBufferFLAT
            self.BRMemorySaverCls = BRMemorySaverFLAT

        else:
            raise ValueError(t_prof.nn_type)

    def reset(self, p_training, eval_opponent_state_dict, stack_size):
        self._rlbr_seat_id = p_training
        self._agent_seat_id = 1 - p_training
        self._opponent = self._eval_agent_cls(t_prof=self._t_prof)
        self._opponent.load_state_dict(eval_opponent_state_dict)
        self._rlbr_env_wrapper = self._eval_env_bldr.get_new_wrapper(is_evaluating=True, stack_size=stack_size)
        self._ddqns[p_training] = DDQN(owner=p_training, ddqn_args=self._args.ddqn_args,
                                       env_bldr=self._eval_env_bldr)
        self._buf = self.CircularBufferCls(env_bldr=self._env_bldr, max_size=self._args.ddqn_args.cir_buf_size)
        self._br_memory_saver = self.BRMemorySaverCls(env_bldr=self._eval_env_bldr, buffer=self._buf)

    def get_grads(self, p_id):
        return self._ray.grads_to_numpy(self._ddqns[p_id].get_grads_one_batch_from_buffer(buffer=self._buf))

    def play(self, n_episodes):
        self._ddqns[self._rlbr_seat_id].eval()
        accumulated_rew = 0.0
        for n in range(n_episodes):

            # """""""""""""""""
            # Reset
            # """""""""""""""""
            obs, r_for_all, done, info = _util.reset_episode_multi_action_space(rlbr_env_wrapper=self._rlbr_env_wrapper,
                                                                                opponent_agent=self._opponent)

            range_idxs = [
                self._rlbr_env_wrapper.env.get_range_idx(p_id=p_id)
                for p_id in range(self._eval_env_bldr.N_SEATS)
            ]

            # Store last game to buffer and reset memory saver
            self._br_memory_saver.reset(range_idx=range_idxs[self._rlbr_seat_id])

            # """""""""""""""""
            # Play Episode
            # """""""""""""""""
            while not done:

                p_id_acting = self._rlbr_env_wrapper.env.current_player.seat_id
                if self._t_prof.DEBUGGING:
                    if p_id_acting != self._opponent._internal_env_wrapper.env.current_player.seat_id:
                        raise RuntimeError("Bad bad bug in RL-BR.")

                # RL-BR acting
                if p_id_acting == self._rlbr_seat_id:
                    legal_actions_list_t = self._rlbr_env_wrapper.env.get_legal_actions()
                    action_int = self._ddqns[self._rlbr_seat_id].select_br_a(
                        pub_obses=[obs],
                        range_idxs=np.array([range_idxs[self._rlbr_seat_id]], dtype=np.int32),
                        explore=True,
                        legal_actions_lists=[legal_actions_list_t]
                    )[0].item()

                    self._br_memory_saver.add_non_terminal_experience(obs_t_before_acted=obs,
                                                                      a_selected_t=action_int,
                                                                      legal_actions_list_t=legal_actions_list_t)

                    # Notify agent
                    _util.notify_agent_multi_action_space(action_int=action_int, rlbr_seat_id=self._rlbr_seat_id,
                                                          rlbr_env_wrapper=self._rlbr_env_wrapper,
                                                          opponent_agent=self._opponent)

                    # Step
                    obs, r_for_all, done, info = self._rlbr_env_wrapper.step(action=action_int)

                # EvalAgent (opponent) acting
                else:
                    action_int, _ = self._opponent.get_action(step_env=True, need_probs=False)
                    obs, r_for_all, done, info = _util.step_from_opp_action(action_int=action_int,
                                                                            opponent=self._opponent,
                                                                            rlbr_env_wrapper=self._rlbr_env_wrapper)

            # Add terminal memory
            self._br_memory_saver.add_terminal(reward_p=r_for_all[self._rlbr_seat_id], terminal_obs=obs)

            # For tracking running reward while training
            accumulated_rew += r_for_all[self._rlbr_seat_id]

        return accumulated_rew \
               * self._eval_env_bldr.env_cls.EV_NORMALIZER \
               * self._rlbr_env_wrapper.env.REWARD_SCALAR \
               / n_episodes

    def update_target_net(self, p_id):
        self._ddqns[p_id].update_target_net()

    def update(self, eps, nets):
        for p_id in range(self._t_prof.n_seats):
            if nets[p_id] is not None:
                self._ddqns[p_id].load_net_state_dict(
                    state_dict=self._ray.state_dict_to_torch(self._ray.get(nets[p_id]),
                                                             device=self._ddqns[p_id].device))

            if eps[p_id] is not None:
                self._ddqns[p_id].eps = self._ray.get(eps[p_id])



# Copyright (c) 2019 Eric Steinberger


import torch

from PokerRL.rl import rl_util


class NetWrapperBase:

    def __init__(self, net, env_bldr, args, owner, device):
        self._env_bldr = env_bldr
        self._args = args
        self.owner = owner
        self.device = device

        self._criterion = rl_util.str_to_loss_cls(self._args.loss_str)
        self.loss_last_batch = None

        self._net = net

    def get_grads_one_batch_from_buffer(self, buffer):
        if buffer.size < self._args.batch_size:
            return

        self._net.train()
        # Might want to do multiple mini batches because all together don't fit into every gpu's vRAM.
        _grad_mngr = _GradManager(net=self._net, args=self._args, criterion=self._criterion)
        for micro_batch_id in range(self._args.n_mini_batches_per_update):
            self._mini_batch_loop(buffer=buffer, grad_mngr=_grad_mngr)

        self.loss_last_batch = _grad_mngr.get_loss_sum()
        return _grad_mngr.average()

    def _mini_batch_loop(self, buffer, grad_mngr):
        raise NotImplementedError

    def load_net_state_dict(self, state_dict):
        self._net.load_state_dict(state_dict)

    def net_state_dict(self):
        return self._net.state_dict()

    def train(self):
        self._net.train()

    def eval(self):
        self._net.eval()

    def state_dict(self):
        """ Override, if necessary"""
        return self.net_state_dict()

    def load_state_dict(self, state):
        """ Override, if necessary"""
        self.load_net_state_dict(state)


class NetWrapperArgsBase:

    def __init__(self,
                 batch_size,
                 n_mini_batches_per_update,
                 optim_str,
                 loss_str,
                 lr,
                 grad_norm_clipping,
                 device_training
                 ):
        assert isinstance(device_training, str), "Please pass a string (either 'cpu' or 'cuda')!"
        self.batch_size = batch_size
        self.n_mini_batches_per_update = n_mini_batches_per_update
        self.optim_str = optim_str
        self.loss_str = loss_str
        self.lr = lr
        self.grad_norm_clipping = grad_norm_clipping
        self.device_training = torch.device(device_training)


class _GradManager:

    def __init__(self, args, net, criterion):
        self.net = net
        self.args = args
        self.criterion = criterion
        self._grads = {}
        self._loss_sum = 0.0
        for name, _ in net.named_parameters():
            self._grads[name] = []

    def backprop(self, pred, target, loss_weights=None):
        self.net.zero_grad()
        if loss_weights is None:
            loss = self.criterion(pred, target)
        else:
            loss = self.criterion(pred, target, loss_weights)
        loss.backward()
        self._loss_sum += loss.item()
        self.add()

    def add(self):
        for name, param in self.net.named_parameters():
            self._grads[name].append(param.grad.data.clone())

    def average(self):
        if self.args.n_mini_batches_per_update == 1:
            for name, param in self.net.named_parameters():
                self._grads[name] = self._grads[name][0]
        else:
            for name, param in self.net.named_parameters():
                self._grads[name] = torch.mean(torch.stack(self._grads[name], dim=0), dim=0)
        return self._grads

    def get_loss_sum(self):
        return self._loss_sum



# Copyright (c) 2019 Eric Steinberger


import copy

from PokerRL.eval._.EvaluatorMasterBase import EvaluatorMasterBase
from PokerRL.game._.tree.PublicTree import PublicTree
from PokerRL.rl import rl_util


class LocalBRMaster(EvaluatorMasterBase):
    """
    (Local) Master to evaluate agents by computing an exact best-response strategy and thereby evaluating the agent's
    exploitability. Note that this type of evaluation should only be used in games with up to a million information sets
    as the provided implementation is not meant for big games. Other evaluators provided are better suited for those.
    """

    def __init__(self, t_prof, chief_handle, eval_agent_cls):
        super().__init__(t_prof=t_prof, eval_env_bldr=rl_util.get_env_builder(t_prof=t_prof), chief_handle=chief_handle,
                         eval_type="BR")
        self._env_bldr = rl_util.get_env_builder(t_prof=t_prof)

        assert self._env_bldr.N_SEATS == 2

        self._eval_agent = eval_agent_cls(t_prof=t_prof)

        self._game_trees = [
            PublicTree(env_bldr=self._env_bldr,
                       stack_size=stack_size,
                       stop_at_street=None,
                       put_out_new_round_after_limit=True,
                       is_debugging=self._t_prof.DEBUGGING)
            for stack_size in self._t_prof.eval_stack_sizes
        ]

        for gt in self._game_trees:
            gt.build_tree()
            print("Tree with stack size", gt.stack_size, "has", gt.n_nodes, "nodes out of which", gt.n_nonterm,
                  "are non-terminal.")

    def evaluate(self, iter_nr):
        for mode in self._t_prof.eval_modes_of_algo:

            if self._is_multi_stack:
                total_of_all_stacks = []

            for stack_size_idx, stack_size in enumerate(self._t_prof.eval_stack_sizes):
                self._eval_agent.set_mode(mode)
                self._eval_agent.set_stack_size(stack_size=stack_size)
                if self._eval_agent.can_compute_mode():
                    expl_p0, expl_p1 = self._compute_br_heads_up(stack_size_idx=stack_size_idx, iter_nr=iter_nr)
                    self._log_results(iter_nr=iter_nr, agent_mode=mode, stack_size_idx=stack_size_idx,
                                      score=(expl_p0 + expl_p1) / 2)

                if self._is_multi_stack:
                    total_of_all_stacks.append((expl_p0 + expl_p1) / 2.0)

            if self._is_multi_stack:
                self._log_multi_stack(agent_mode=mode,
                                      iter_nr=iter_nr,
                                      score_total=sum(total_of_all_stacks) / float(len(total_of_all_stacks)))

    def update_weights(self):
        w = self.pull_current_strat_from_chief()
        self._eval_agent.update_weights(copy.deepcopy(w))

    def _compute_br_heads_up(self, stack_size_idx, iter_nr=None, do_export_tree=True):
        gt = self._game_trees[stack_size_idx]
        gt.fill_with_agent_policy(agent=self._eval_agent)
        gt.compute_ev()

        if do_export_tree:
            gt.export_to_file(name=self._t_prof.name + "__BR_vs_" + self._eval_agent.get_mode()
                                   + "__stack_idx"
                                   + str(stack_size_idx) + "_I_" + str(iter_nr))

        expl_p1 = float(gt.root.exploitability[0]) * self._env_bldr.env_cls.EV_NORMALIZER
        expl_p2 = float(gt.root.exploitability[1]) * self._env_bldr.env_cls.EV_NORMALIZER

        return expl_p1, expl_p2



# Copyright (c) 2019 Eric Steinberger


import copy

from PokerRL.game import bet_sets
from PokerRL.game.poker_env_args import DiscretizedPokerEnvArgs, LimitPokerEnvArgs, NoLimitPokerEnvArgs


class LBRArgs:
    """
    Argument object for LBR
    """

    def __init__(self,
                 lbr_bet_set=bet_sets.OFF_TREE_11,
                 n_lbr_hands_per_seat=30000,
                 lbr_check_to_round=None,  # recommended to set to Poker.TURN for 4-round games.
                 n_parallel_lbr_workers=10,
                 use_gpu_for_batch_eval=True,
                 DISTRIBUTED=False,
                 ):
        """

        Args:
            lbr_bet_set (list):             List of floats; fractions of pot that LBR can consider to bet. (Note:
                                            only relevant in discretized and no-limit games. Limit games always allow 3
                                            actions)
            n_lbr_hands_per_seat (int):              Number of LBR hands to compute.

            lbr_check_to_round (game round from Poker):
                                            In the original paper, LBR has been shown to perform better, when it only
                                            check/calls until the Turn in No-Limit Texas Hold'em. It is also faster if
                                            one does that. We recommend setting this to Poker.TURN in games with 4
                                            rounds and to None in Leduc-like games.
            n_parallel_lbr_workers (int):   Number of workers. Only relevant if running distributed
            use_gpu_for_batch_eval (bool):  Whether to use the GPU for batched strategy queries. Recommended for
                                            big neural networks and games with many different hands (like Hold'em), if
                                            available.
            DISTRIBUTED (bool):             Whether to use ray and run distributed or to run locally.
        """
        self.lbr_bet_set = lbr_bet_set
        self.n_lbr_hands = n_lbr_hands_per_seat
        self.lbr_check_to_round = lbr_check_to_round
        self.n_workers = n_parallel_lbr_workers if DISTRIBUTED else 1
        self.use_gpu_for_batch_eval = use_gpu_for_batch_eval
        self.DISTRIBUTED = DISTRIBUTED

    def get_lbr_env_args(self, agents_env_args):
        arg_cls = type(agents_env_args)

        if arg_cls is DiscretizedPokerEnvArgs:
            return DiscretizedPokerEnvArgs(
                n_seats=agents_env_args.n_seats,
                starting_stack_sizes_list=copy.deepcopy(agents_env_args.starting_stack_sizes_list),
                bet_sizes_list_as_frac_of_pot=copy.deepcopy(self.lbr_bet_set),
                stack_randomization_range=(0, 0),
                use_simplified_headsup_obs=agents_env_args.use_simplified_headsup_obs,
                uniform_action_interpolation=False
            )

        elif arg_cls is LimitPokerEnvArgs:
            return LimitPokerEnvArgs(
                n_seats=agents_env_args.n_seats,
                starting_stack_sizes_list=copy.deepcopy(agents_env_args.starting_stack_sizes_list),
                stack_randomization_range=(0, 0),
                use_simplified_headsup_obs=agents_env_args.use_simplified_headsup_obs,
                uniform_action_interpolation=False
            )

        elif arg_cls is NoLimitPokerEnvArgs:
            raise NotImplementedError("Currently not supported")

        else:
            raise TypeError(arg_cls)



# Copyright (c) 2019 Eric Steinberger
import copy

import numpy as np


class InteractiveGame:
    """
    This class facilitates play between a user against an EvalAgent or against himself in any poker game.
    """

    def __init__(self, env_cls, env_args, seats_human_plays_list, eval_agent=None):
        """

        Args:
            env_cls (PokerEnv subclass):    the subclass of PokerEnv (not instance!) that the EvalAgent was trained on
                                            or more generally, you want to play in.

            env_args (PokerEnvArgs, DiscretizedPokerEnvArgs, or LimitPokerEnvArgs):
                                            The arguments object corresponding to the environment

            seats_human_plays_list (list):  a list of ints that indicates which seats on the table are played by the
                                            human in the command line. If you pass an empty list, you can watch the AI
                                            play against itself, or you can pass a list with all seats to play against
                                            yourself. The most common use-case, however, is to pass some seats and
                                            play against the agent.

            eval_agent (EvalAgentBase):     The wrapped agent. You only need to pass one if ""seats_human_plays_list""
                                            doesn't cover all seats on the table.
        """
        if len(seats_human_plays_list) < env_args.n_seats:
            assert eval_agent is not None

        self._env = env_cls(env_args=env_args, is_evaluating=True, lut_holder=env_cls.get_lut_holder())
        self._env.reset()

        self._eval_agent = eval_agent

        self._seats_human_plays_list = seats_human_plays_list
        self._winnings_per_seat = [0 for _ in range(self._env.N_SEATS)]

    @property
    def seats_human_plays_list(self):
        return copy.deepcopy(self._seats_human_plays_list)

    @property
    def winnings_per_seat(self):
        return copy.deepcopy(self._winnings_per_seat)

    def start_to_play(self, render_mode="TEXT", limit_numpy_digits=True):
        if limit_numpy_digits:
            np.set_printoptions(precision=5, suppress=True)
        print("""                       
                                                _____
                    _____                _____ |6    |
                   |2    | _____        |5    || & & | 
                   |  &  ||3    | _____ | & & || & & | _____
                   |     || & & ||4    ||  &  || & & ||7    |
                   |  &  ||     || & & || & & ||____9|| & & | _____
                   |____Z||  &  ||     ||____S|       |& & &||8    | _____
                          |____E|| & & |              | & & ||& & &||9    |
                                 |____h|              |____L|| & & ||& & &|
                                                             |& & &||& & &|
                                                             |____8||& & &|
                                                                    |____6|
               """)
        self._env.print_tutorial()

        # play forever until human player manually stops
        while True:
            print()
            print("****************************")
            print("*        GAME START        *")
            print("****************************")
            print()

            # ______________________________________________ one episode _______________________________________________

            self._env.reset()
            if self._eval_agent is not None:
                # The agent has his own copy of the env within its EvalAgent. The following lines sets the decks between
                # self.env and the agent's env equal. However, the agent cannot see any private cards but his own in his
                # observations, of course!
                self._eval_agent.reset(deck_state_dict=self._env.cards_state_dict())

            self._env.render(mode=render_mode)
            while True:
                current_player_id = self._env.current_player.seat_id

                if self._eval_agent is not None:
                    assert np.array_equal(self._env.board, self._eval_agent._internal_env_wrapper.env.board)
                    assert np.array_equal(np.array(self._env.side_pots),
                                          np.array(self._eval_agent._internal_env_wrapper.env.side_pots))
                    assert self._env.current_player.seat_id == \
                           self._eval_agent._internal_env_wrapper.env.current_player.seat_id
                    assert self._env.current_round == self._eval_agent._internal_env_wrapper.env.current_round

                # Human acts
                if current_player_id in self._seats_human_plays_list:
                    action_tuple = self._env.human_api_ask_action()

                    if self._eval_agent is not None:
                        self._eval_agent.notify_of_processed_tuple_action(action_tuple=action_tuple,
                                                                          p_id_acted=current_player_id)

                # Agent acts
                else:
                    a_idx, frac = self._eval_agent.get_action_frac_tuple(step_env=True)
                    if a_idx == 2:
                        action_tuple = [2, self._env.get_fraction_of_pot_raise(fraction=frac,
                                                                               player_that_bets=current_player_id)]
                    else:
                        action_tuple = [a_idx, -1]

                obs, rews, done, info = self._env._step(processed_action=action_tuple)
                self._env.render(mode=render_mode)

                if done:
                    break

            for s in range(self._env.N_SEATS):
                self._winnings_per_seat[s] += np.rint(rews[s] * self._env.REWARD_SCALAR)

            print("")
            print("Current Winnings per player:", self._winnings_per_seat)
            input("Press Enter to go to the next round.")



# Copyright (c) 2019 Eric Steinberger


"""
Wraps the local LBR worker in a ray actor to be placed on any machine in the cluster. You can spawn as many of these
as you want to accelerate the LBR computation; the EvalLBRMaster will manage them all.
"""

import ray
import torch

from PokerRL.eval.lbr.LocalLBRWorker import LocalLBRWorker as LocalEvalLBRWorker


@ray.remote(num_cpus=1, num_gpus=1 if torch.cuda.is_available() else 0)
class DistLBRWorker(LocalEvalLBRWorker):

    def __init__(self, t_prof, chief_handle, eval_agent_cls):
        LocalEvalLBRWorker.__init__(self, t_prof=t_prof, chief_handle=chief_handle, eval_agent_cls=eval_agent_cls)



# Copyright (c) 2019 Eric Steinberger


import numpy as np

from PokerRL.rl.base_cls.workers.WorkerBase import WorkerBase


class EvaluatorMasterBase(WorkerBase):
    """
    Baseclass to all Evaluators. An Evaluator is an algorithm to evaluate an agent's performance in a certain metric.
    """

    def __init__(self, t_prof, eval_env_bldr, chief_handle, eval_type, log_conf_interval=False):
        """
        Args:
            t_prof (TrainingProfile)
            chief_handle (class instance or ray ActorHandle)
            eval_type (str):                Name of the evaluator
        """
        super().__init__(t_prof=t_prof)
        self._eval_env_bldr = eval_env_bldr
        self._chief_handle = chief_handle
        self._chief_info = [None for _ in range(self._t_prof.n_seats)]

        self._is_multi_stack = len(self._t_prof.eval_stack_sizes) > 1

        self._log_conf_interval = log_conf_interval

        self._exp_name_total, self._exp_names_conf = self._create_experiments(self_name=eval_type)

        if self._is_multi_stack:
            self._exp_name_multi_stack = {
                eval_mode:
                    self._ray.get(
                        self._ray.remote(self._chief_handle.create_experiment,
                                         self._t_prof.name
                                         + " " + eval_mode
                                         + "Multi_Stack"
                                         + ": " + eval_type
                                         + " Averaged Total"))
                for eval_mode in self._t_prof.eval_modes_of_algo
            }
            if self._log_conf_interval:
                self._exp_names_multi_stack_conf = {
                    eval_mode:
                        self._ray.get(
                            [
                                self._ray.remote(self._chief_handle.create_experiment,
                                                 self._t_prof.name
                                                 + " " + eval_mode
                                                 + ": " + eval_type
                                                 + " Conf_" + bound_end)
                                for bound_end in ["lower95", "upper95"]
                            ]
                        )
                    for eval_mode in self._t_prof.eval_modes_of_algo
                }

    @property
    def is_multi_stack(self):
        """
        Whether the agent is evaluated in games that start with different stack sizes each time.
        """
        return self._is_multi_stack

    def evaluate(self, iter_nr):
        """ Evaluate an agent and send the results as logs to the Chief. """
        raise NotImplementedError

    def update_weights(self):
        """ Update the local weights on the master, for instance by calling .pull_current_strat_from_chief()  """
        raise NotImplementedError

    def pull_current_strat_from_chief(self):
        """
        Pulls and Returns weights or any other changing algorithm info of any format from the Chief.
        """
        w, self._chief_info = self._ray.get(self._ray.remote(self._chief_handle.pull_current_eval_strategy,
                                                             self._chief_info))
        return w

    def _create_experiments(self, self_name, ):
        """
        Registers a new experiment either for each player and their average or just for their average.
        """

        if self._log_conf_interval:
            exp_names_conf = {
                eval_mode:
                    [
                        self._ray.get(
                            [
                                self._ray.remote(self._chief_handle.create_experiment,
                                                 self._t_prof.name
                                                 + " " + eval_mode
                                                 + "_stack_" + str(stack_size[0])
                                                 + ": " + self_name
                                                 + " Conf_" + bound_end)
                                for bound_end in ["lower95", "upper95"]
                            ]
                        )
                        for stack_size in self._t_prof.eval_stack_sizes
                    ]
                for eval_mode in self._t_prof.eval_modes_of_algo
            }
        else:
            exp_names_conf = None

        exp_name_total = {
            eval_mode:
                [
                    self._ray.get(
                        self._ray.remote(self._chief_handle.create_experiment,
                                         self._t_prof.name
                                         + " " + eval_mode
                                         + "_stack_" + str(stack_size[0])
                                         + ": " + self_name
                                         + " Total"))
                    for stack_size in self._t_prof.eval_stack_sizes
                ]
            for eval_mode in self._t_prof.eval_modes_of_algo
        }

        return exp_name_total, exp_names_conf

    def _get_95confidence(self, scores):
        mean = np.mean(scores).item()
        std = np.std(scores).item()

        _d = 1.96 * std / np.sqrt(scores.shape[0])
        return float(mean), float(_d)

    def _log_results(self, agent_mode, stack_size_idx, iter_nr, score, upper_conf95=None, lower_conf95=None):
        """
        Log evaluation results by sending these results to the Chief, who will later send them to the Crayon log server.

        Args:
            agent_mode:             Evaluation mode of the agent whose performance is logged
            stack_size_idx:         If evaluating multiple starting stack sizes, this is an index describing which one
                                    this data is from.
            iter_nr:                Algorithm Iteration of this data
            score:                  Score in this evaluation (e.g. exploitability)
        """
        graph_name = "Evaluation/" + self._eval_env_bldr.env_cls.WIN_METRIC

        self._ray.remote(self._chief_handle.add_scalar,
                         self._exp_name_total[agent_mode][stack_size_idx], graph_name, iter_nr, score)

        if self._log_conf_interval:
            assert upper_conf95 is not None
            assert lower_conf95 is not None
            self._ray.remote(self._chief_handle.add_scalar,
                             self._exp_names_conf[agent_mode][stack_size_idx][0], graph_name, iter_nr, lower_conf95)
            self._ray.remote(self._chief_handle.add_scalar,
                             self._exp_names_conf[agent_mode][stack_size_idx][1], graph_name, iter_nr, upper_conf95)

    def _log_multi_stack(self, agent_mode, iter_nr, score_total, upper_conf95=None, lower_conf95=None):
        """
        Additional logging for multistack evaluations
        """
        graph_name = "Evaluation/" + self._eval_env_bldr.env_cls.WIN_METRIC
        self._ray.remote(self._chief_handle.add_scalar,
                         self._exp_name_multi_stack[agent_mode], graph_name, iter_nr, score_total)

        if self._log_conf_interval:
            assert upper_conf95 is not None
            assert lower_conf95 is not None
            self._ray.remote(self._chief_handle.add_scalar,
                             self._exp_names_multi_stack_conf[agent_mode][0], graph_name, iter_nr, lower_conf95)
            self._ray.remote(self._chief_handle.add_scalar,
                             self._exp_names_multi_stack_conf[agent_mode][1], graph_name, iter_nr, upper_conf95)



from .LBRArgs import *
from .LocalLBRMaster import *
from .LocalLBRWorker import *

try:
    import ray
    from .DistLBRMaster import *
    from .DistLBRWorker import *

except ImportError:
    pass



# Copyright (c) 2019 Eric Steinberger


from PokerRL.game._.rl_env.base.PokerEnv import PokerEnv


class Wrapper:

    def __init__(self, env, env_bldr_that_built_me):
        """
        Args:
            env (PokerEnv subclass instance):   The environment instance to be wrapped

            env_bldr_that_built_me:          EnvWrappers should only be created by EnvBuilders. The EnvBuilder
                                                instance passes ""self"" as the value for this argument.
        """
        assert issubclass(type(env), PokerEnv)
        self.env = env
        self.env_bldr = env_bldr_that_built_me

    def _return_obs(self, rew_for_all_players, done, info, env_obs=None):
        return self.get_current_obs(env_obs=env_obs), rew_for_all_players, done, info

    # _______________________________ directly interact with the env inside the wrapper ________________________________
    def step(self, action):
        """
        Steps the environment from an action of the natural action representation to the environment.

        Returns:
            obs, reward, done, info
        """
        env_obs, rew_for_all_players, done, info = self.env.step(action)
        self._pushback(env_obs)
        return self._return_obs(env_obs=env_obs, rew_for_all_players=rew_for_all_players, done=done, info=info)

    def step_from_processed_tuple(self, action):
        """
        Steps the environment from a tuple (action, num_chips,).

        Returns:
            obs, reward, done, info
        """
        env_obs, rew_for_all_players, done, info = self.env.step_from_processed_tuple(action)
        self._pushback(env_obs)
        return self._return_obs(env_obs=env_obs, rew_for_all_players=rew_for_all_players, done=done, info=info)

    def step_raise_pot_frac(self, pot_frac):
        """
        Steps the environment from a fractional pot raise instead of an action as usually specified.

        Returns:
            obs, reward, done, info
        """
        env_obs, rew_for_all_players, done, info = self.env.step_raise_pot_frac(pot_frac=pot_frac)
        self._pushback(env_obs)
        return self._return_obs(env_obs=env_obs, rew_for_all_players=rew_for_all_players, done=done, info=info)

    def reset(self, deck_state_dict=None):
        env_obs, rew_for_all_players, done, info = self.env.reset(deck_state_dict=deck_state_dict)
        self._reset_state()
        self._pushback(env_obs)
        return self._return_obs(env_obs=env_obs, rew_for_all_players=rew_for_all_players, done=done, info=info)

    # ___________________________________________________ To Override __________________________________________________
    def state_dict(self):
        return {"env": self.env.state_dict()}

    def load_state_dict(self, state_dict):
        self.env.load_state_dict(state_dict["env"])

    def _reset_state(self):
        raise NotImplementedError

    def _pushback(self, env_obs):
        """
        Processes a transition in the wrapper. This should be called before every action by any agent.
        """
        raise NotImplementedError

    def get_current_obs(self, env_obs):
        raise NotImplementedError

    def print_obs(self, wrapped_obs):
        raise NotImplementedError

    def set_to_public_tree_node_state(self, node):
        raise NotImplementedError



# Copyright (c) 2019 Eric Steinberger


import copy

import numpy as np


class CircularBufferBase:
    """
    self.games stores references to Game subclass objects. One Game instance might be referenced multiple times,
    depending on the number of steps that it contains. This is to keep equally likely sampling.
    """

    def __init__(self, env_bldr, max_size):
        self._env_bldr = env_bldr
        self._max_size = max_size
        self._size = None
        self._top = None

    def sample(self, device, batch_size):
        raise NotImplementedError

    def state_dict(self):
        raise NotImplementedError

    def load_state_dict(self, state):
        raise NotImplementedError

    def reset(self):
        self._size = 0
        self._top = 0


class BRMemorySaverBase:
    """ Interface for correct BR reward storing """

    def __init__(self, env_bldr, buffer):
        self._env_bldr = env_bldr
        self._buffer = buffer
        self._intermediate_memory = _PlayerTimeStepMemory()

    def add_terminal(self,
                     reward_p,
                     terminal_obs,
                     ):
        raise NotImplementedError

    def add_non_terminal_experience(self,
                                    obs_t_before_acted,
                                    a_selected_t,
                                    legal_actions_list_t):
        raise NotImplementedError

    def _add_step_to_memory(self):
        raise NotImplementedError

    def reset(self, range_idx):
        raise NotImplementedError


class _PlayerTimeStepMemory:

    def __init__(self):
        self.obs_t = None
        self.obs_tp1 = None
        self.action = None
        self.action_tp1 = None
        self.legal_actions_list_t = None
        self.legal_actions_list_tp1 = None

    def add_experience(self, obs_t, action_t, legal_actions_list_t):
        """
        called when it is player's turn.
        """
        if self.obs_t is None:
            self.obs_t = np.copy(obs_t)
            self.action = action_t
            self.legal_actions_list_t = np.copy(legal_actions_list_t)
        elif self.obs_tp1 is None:
            self.obs_tp1 = np.copy(obs_t)
            self.action_tp1 = action_t
            self.legal_actions_list_tp1 = np.copy(legal_actions_list_t)
        else:
            raise BufferError("This should not happen.")

    def is_level_1(self):
        return self.obs_t is not None and self.obs_tp1 is None

    def is_complete(self):
        return self.obs_tp1 is not None

    def step(self):
        self.obs_t = np.copy(self.obs_tp1)
        self.obs_tp1 = None
        self.action = self.action_tp1
        self.action_tp1 = None
        self.legal_actions_list_t = copy.deepcopy(self.legal_actions_list_tp1)
        self.legal_actions_list_tp1 = None

    def reset(self):
        self.obs_t = None
        self.obs_tp1 = None
        self.action = None
        self.action_tp1 = None
        self.legal_actions_list_t = None
        self.legal_actions_list_tp1 = None



# Copyright (c) 2019 Eric Steinberger


import torch
import torch.nn as nn


class AvrgStrategyNet(nn.Module):

    def __init__(self, avrg_net_args, env_bldr, device):
        super().__init__()
        self.args = avrg_net_args
        self.env_bldr = env_bldr
        self.n_actions = self.env_bldr.N_ACTIONS

        MPM = avrg_net_args.mpm_args.get_mpm_cls()

        self._relu = nn.ReLU(inplace=False)
        self._mpm = MPM(env_bldr=env_bldr, device=device, mpm_args=self.args.mpm_args)

        self._final_layer = nn.Linear(in_features=self._mpm.output_units, out_features=self.args.n_units_final)
        self._out_layer = nn.Linear(in_features=self.args.n_units_final, out_features=self.n_actions)

        self.to(device)

    def forward(self, pub_obses, range_idxs, legal_action_masks):
        """
        Softmax is not applied in here! It is separate in training and action fns
        """

        out = self._mpm(pub_obses=pub_obses, range_idxs=range_idxs)
        out = self._relu(self._final_layer(out))
        out = self._out_layer(out)
        out = torch.where(legal_action_masks == 1,
                          out,
                          torch.FloatTensor([-10e20]).to(device=out.device).expand_as(out))
        return out


class AvrgNetArgs:

    def __init__(self,
                 mpm_args,
                 n_units_final
                 ):
        self.mpm_args = mpm_args
        self.n_units_final = n_units_final



# Copyright (c) 2019 Eric Steinberger
# Inspiration of architecture from DeepStack-Leduc (https://github.com/lifrordi/DeepStack-Leduc/tree/master/Source)

import numpy as np

from PokerRL.game.Poker import Poker
from PokerRL.game.PokerEnvStateDictEnums import EnvDictIdxs
from PokerRL.game._.tree._.nodes import PlayerActionNode


class ValueFiller:

    def __init__(self, tree):
        self._tree = tree
        self._env_bldr = tree.env_bldr
        self._env = self._env_bldr.get_new_env(is_evaluating=True)

        # This only works for 1-Card games!
        self._eq_const = (self._env_bldr.rules.N_CARDS_IN_DECK / (self._env_bldr.rules.N_CARDS_IN_DECK - 1))

    def compute_cf_values_heads_up(self, node):
        """
        The functionality is extremely simplified compared to n-agent evaluations and made for HU Leduc only!
        Furthermore, this BR implementation is *VERY* inefficient and not suitable for anything much bigger than Leduc.
        """

        assert self._tree.n_seats == 2

        if node.is_terminal:
            assert node.strategy is None
        else:
            assert node.strategy.shape == (self._env_bldr.rules.RANGE_SIZE, len(node.children),)

        if node.is_terminal:
            """
            equity: -1*reach=always lose. 1*reach=always win. 0=50%/50%
            """
            assert isinstance(node, PlayerActionNode)

            # Fold
            if node.action == Poker.FOLD:
                if node.env_state[EnvDictIdxs.current_round] == Poker.FLOP:
                    equity = self._get_fold_eq_final_street(node=node)
                else:
                    equity = self._get_fold_eq_preflop(node=node)

            # Check / Call
            else:
                if node.env_state[EnvDictIdxs.current_round] == Poker.FLOP:
                    equity = self._get_call_eq_final_street(reach_probs=node.reach_probs,
                                                            board_2d=node.env_state[EnvDictIdxs.board_2d])

                else:  # preflop
                    equity = self._get_call_eq_preflop(node=node)

            # set boardcards to 0
            for c in self._env_bldr.lut_holder.get_1d_cards(node.env_state[EnvDictIdxs.board_2d]):
                if c != Poker.CARD_NOT_DEALT_TOKEN_1D:
                    equity[:, c] = 0.0

            node.ev = equity * node.env_state[EnvDictIdxs.main_pot] / 2
            node.ev_br = np.copy(node.ev)

        else:
            N_ACTIONS = len(node.children)
            ev_all_actions = np.zeros(shape=(N_ACTIONS, self._tree.n_seats, self._env_bldr.rules.RANGE_SIZE),
                                      dtype=np.float32)
            ev_br_all_actions = np.zeros(shape=(N_ACTIONS, self._tree.n_seats, self._env_bldr.rules.RANGE_SIZE),
                                         dtype=np.float32)

            for i, child in enumerate(node.children):
                self.compute_cf_values_heads_up(node=child)
                ev_all_actions[i] = child.ev
                ev_br_all_actions[i] = child.ev_br

            if node.p_id_acting_next == self._tree.CHANCE_ID:
                node.ev = np.sum(ev_all_actions, axis=0)
                node.ev_br = np.sum(ev_br_all_actions, axis=0)

            else:
                node.ev = np.zeros(shape=(self._tree.n_seats, self._env_bldr.rules.RANGE_SIZE), dtype=np.float32)
                node.ev_br = np.zeros(shape=(self._tree.n_seats, self._env_bldr.rules.RANGE_SIZE), dtype=np.float32)

                plyr = node.p_id_acting_next
                opp = 1 - node.p_id_acting_next

                node.ev[plyr] = np.sum(node.strategy.T * ev_all_actions[:, plyr], axis=0)
                node.ev[opp] = np.sum(ev_all_actions[:, opp], axis=0)

                node.ev_br[opp] = np.sum(ev_br_all_actions[:, opp], axis=0)
                node.ev_br[plyr] = np.max(ev_br_all_actions[:, plyr], axis=0)

                node.br_a_idx_in_child_arr_for_each_hand = np.argmax(ev_br_all_actions[:, plyr], axis=0)

        # weight ev by reach prob
        node.ev_weighted = node.ev * node.reach_probs
        node.ev_br_weighted = node.ev_br * node.reach_probs
        assert np.allclose(np.sum(node.ev_weighted), 0, atol=0.001), np.sum(node.ev_weighted)  # Zero Sum check

        node.epsilon = node.ev_br_weighted - node.ev_weighted
        node.exploitability = np.sum(node.epsilon, axis=1)

    def _get_fold_eq_preflop(self, node):
        equity = np.zeros(shape=(self._tree.n_seats, self._env_bldr.rules.RANGE_SIZE), dtype=np.float32)

        for p in range(self._tree.n_seats):
            opp = 1 - p

            # sum reach probs for all hands and subtracts the reach prob of the hand player p holds batched for all
            equity[p] = np.sum(node.reach_probs[opp]) - node.reach_probs[opp]

        equity[node.p_id_acted_last] *= -1
        return equity * self._eq_const

    def _get_fold_eq_final_street(self, node):
        equity = np.zeros(shape=(self._tree.n_seats, self._env_bldr.rules.RANGE_SIZE), dtype=np.float32)

        for p in range(self._tree.n_seats):
            opp = 1 - p

            # sum reach probs for all hands and subtracts the reach prob of the hand player p holds batched for all
            equity[p] = np.sum(node.reach_probs[opp]) - node.reach_probs[opp]

        equity[node.p_id_acted_last] *= -1
        return equity * self._eq_const

    def _get_call_eq_final_street(self, reach_probs, board_2d):
        """
        Returns:
            equity: negative=lose. positive=win. 0=50%/50%

        """
        c = self._env_bldr.lut_holder.get_1d_cards(board_2d)[0]

        assert c != Poker.CARD_NOT_DEALT_TOKEN_1D

        equity = np.zeros(shape=(self._tree.n_seats, self._env_bldr.rules.RANGE_SIZE), dtype=np.float32)
        handranks = np.empty(shape=self._env_bldr.rules.RANGE_SIZE, dtype=np.int32)

        for h in range(self._env_bldr.rules.RANGE_SIZE):
            handranks[h] = self._env.get_hand_rank(
                board_2d=board_2d,
                hand_2d=self._env_bldr.lut_holder.get_2d_hole_cards_from_range_idx(range_idx=h))

        for p in range(self._tree.n_seats):
            opp = 1 - p
            for h in range(self._env_bldr.rules.RANGE_SIZE):
                if h != c:
                    for h_opp in range(self._env_bldr.rules.RANGE_SIZE):
                        if h_opp != h and h_opp != c:
                            # when same handrank, would be += 0
                            if handranks[h] > handranks[h_opp]:
                                equity[p, h] += reach_probs[opp, h_opp]
                            elif handranks[h] < handranks[h_opp]:
                                equity[p, h] -= reach_probs[opp, h_opp]

        assert np.allclose(equity[:, c], 0)
        return equity * self._eq_const

    def _get_call_eq_preflop(self, node):
        # very Leduc specific
        equity = np.zeros(shape=(self._tree.n_seats, self._env_bldr.rules.RANGE_SIZE), dtype=np.float32)

        for c in range(self._env_bldr.rules.N_CARDS_IN_DECK):
            """ ._get_call_eq() returns 0 for blocked hands, so we are summing 5 hands for each board. """
            _board_1d = np.array([c], dtype=np.int8)
            _board_2d = self._env_bldr.lut_holder.get_2d_cards(_board_1d)
            _reach_probs = np.copy(node.reach_probs)
            _reach_probs[:, c] = 0

            equity += self._get_call_eq_final_street(reach_probs=_reach_probs, board_2d=_board_2d)

        # mean :: using (N_CARDS_IN_DECK - 2) because two boards are blocked (by agent's and opp's cards)
        equity /= (self._env_bldr.rules.N_CARDS_IN_DECK - 2)
        return equity  # * self._eq_const



# Copyright (c) 2019 Eric Steinberger


import unittest
from unittest import TestCase

from PokerRL.cfr.CFRPlus import CFRPlus
from PokerRL.cfr.LinearCFR import LinearCFR
from PokerRL.cfr.VanillaCFR import VanillaCFR
from PokerRL.game import bet_sets
from PokerRL.game.games import DiscretizedNLLeduc
from PokerRL.rl.base_cls.workers.ChiefBase import ChiefBase


class TestCFR(TestCase):

    def test_run_CFR(self):
        n_iterations = 3
        name = "TESTING_CFR"

        chief = ChiefBase(t_prof=None)
        cfr = VanillaCFR(name=name,
                         game_cls=DiscretizedNLLeduc,
                         agent_bet_set=bet_sets.POT_ONLY,
                         chief_handle=chief)

        cfr.reset()

        for iter_id in range(n_iterations):
            cfr.iteration()

    def test_run_CFRplus(self):
        n_iterations = 3
        name = "TESTING_CFRplus"

        chief = ChiefBase(t_prof=None)
        cfr = CFRPlus(name=name,
                      game_cls=DiscretizedNLLeduc,
                      delay=0,
                      agent_bet_set=bet_sets.POT_ONLY,
                      chief_handle=chief)

        cfr.reset()

        for iter_id in range(n_iterations):
            cfr.iteration()

    def test_run_linearCFR(self):

        n_iterations = 3
        name = "TESTING_LinearCFR"

        chief = ChiefBase(t_prof=None)
        cfr = LinearCFR(name=name,
                        game_cls=DiscretizedNLLeduc,
                        agent_bet_set=bet_sets.POT_ONLY,
                        chief_handle=chief)

        cfr.reset()

        for iter_id in range(n_iterations):
            cfr.iteration()


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Eric Steinberger


"""
Wraps the local H2H master in a ray actor (i.e. worker) to be placed on any (one) machine in the cluster.
The H2H computation itself is not distributed.
"""

import ray
import torch

from PokerRL.eval.head_to_head.LocalHead2HeadMaster import LocalHead2HeadMaster as LocalEvalHead2HeadMaster


@ray.remote(num_cpus=1, num_gpus=1 if torch.cuda.is_available() else 0)
class DistHead2HeadMaster(LocalEvalHead2HeadMaster):

    def __init__(self, t_prof, chief_handle, eval_agent_cls):
        LocalEvalHead2HeadMaster.__init__(self, t_prof=t_prof, chief_handle=chief_handle, eval_agent_cls=eval_agent_cls)



# Copyright (c) 2019 Eric Steinberger


"""
Enum keys to PokerEnv's state_dict objects.
"""


class EnvDictIdxs:
    current_round = "current_round"
    side_pots = "side_pots"
    main_pot = "main_pot"
    board_2d = "board_2d"
    last_action = "last_action"
    capped_raise = "capped_raise"
    current_player = "current_player"
    last_raiser = "last_raiser"
    deck = "deck_remaining"
    seats = "seats"
    n_actions_this_episode = "n_actions_this_episode"
    # only relevant for Fixed-Limit games
    n_raises_this_round = "n_raises_this_round"

    is_evaluating = "is_evaluating"


class PlayerDictIdxs:
    hand = "hand"
    hand_rank = "hand_rank"
    stack = "stack"
    current_bet = "current_bet"
    is_allin = "is_allin"
    folded_this_episode = "folded_this_episode"
    has_acted_this_round = "has_acted_this_round"
    side_pot_rank = "side_pot_rank"
    seat_id = "seat_id"



# Copyright (c) 2019 Eric Steinberger


"""
This script will start a game of 3-player No-Limit Texas Hold'em with discrete bet sizes in which you can play against
yourself.
"""

from PokerRL.game.InteractiveGame import InteractiveGame
from PokerRL.game.games import DiscretizedNLHoldem

if __name__ == '__main__':
    game_cls = DiscretizedNLHoldem
    args = game_cls.ARGS_CLS(n_seats=3,
                             bet_sizes_list_as_frac_of_pot=[
                                 0.2,
                                 0.5,
                                 1.0,
                                 2.0,
                                 1000.0  # Note that 1000x pot will always be >pot and thereby represents all-in
                             ],
                             stack_randomization_range=(0, 0,),
                             )

    game = InteractiveGame(env_cls=game_cls,
                           env_args=args,
                           seats_human_plays_list=[0, 1, 2],
                           )

    game.start_to_play()



# Copyright (c) 2019 Eric Steinberger
import numpy as np
import psutil
import torch

# not being able to import ray is okay if you don't run distributed!
try:
    import ray
    import ray.utils
except ModuleNotFoundError:
    pass


class MaybeRay:
    """
    These ray wrapper functions serve to unify the distributed and local function call syntax. If you call them,
    but do not run distributed, everything will just be tunneled through and ray will not even be imported,
    but the same code can be run distributed just by enabling a single flag ""runs_distributed"".
    """

    def __init__(self, runs_distributed=True, runs_cluster=False):
        """
        Args:k
            runs_distributed (bool): all ray calls are skipped if this is false
        """
        if runs_cluster:
            assert runs_distributed
        self.runs_distributed = runs_distributed
        self.runs_cluster = runs_cluster

    # __________________________________________________ ray wrapper ___________________________________________________
    def init_cluster(self, redis_address):
        assert self.runs_cluster
        ray.init(
            redis_address=redis_address,
            redis_max_memory=min(10 ** 10, int(psutil.virtual_memory().total * 0.1)),
            object_store_memory=min(2 * (10 ** 10), int(psutil.virtual_memory().total * 0.4)),
        )

    def init_local(self):
        if self.runs_distributed:
            ray.init(
                redis_max_memory=min(10 ** 10, int(psutil.virtual_memory().total * 0.1)),
                object_store_memory=min(2 * (10 ** 10), int(psutil.virtual_memory().total * 0.4)),
            )

    def get(self, obj):
        if self.runs_distributed:
            return ray.get(obj)
        return obj

    def put(self, obj):
        if self.runs_distributed:
            return ray.put(obj)
        return obj

    def remote(self, fn, *args):
        if self.runs_distributed:
            return fn.remote(*args)
        return fn(*args)

    def create_worker(self, cls, *args):
        if self.runs_distributed:
            return cls.remote(*args)
        return cls(*args)

    def wait(self, _list, num_returns=None, timeout=None, return_not_ready=False):
        """
        Args:
            _list (list:                    list of object ids to wait for
            num_returns (int):              if None: wait for all; if any number waits for that number
            timeout (int):                  Optional. If specified, waits only for ""timeout"" milliseconds
            return_not_ready (bool):        if True: returns tuple (rdy, not_rdy). If false: returns rdy
        """
        if self.runs_distributed:
            num_returns = len(_list) if num_returns is None else num_returns
            rdy, not_rdy = ray.wait(_list, num_returns=num_returns, timeout=timeout)
            if return_not_ready:
                return rdy, not_rdy
            return rdy

        if return_not_ready:
            return _list, []
        return _list

    # __________________________________________________ util tools ____________________________________________________
    def state_dict_to_numpy(self, _dict):
        """ if local, skip torch <--> numpy """
        if self.runs_distributed:
            np_dict = {}
            for k in list(_dict.keys()):
                if isinstance(_dict[k], np.ndarray):
                    np_dict[k] = _dict[k]
                else:
                    np_dict[k] = _dict[k].cpu().numpy()

            return np_dict

        return _dict

    def state_dict_to_torch(self, _dict, device):
        """ if local, skip torch <--> numpy """

        new_dict = {}
        if self.runs_distributed:
            for k in list(_dict.keys()):
                if isinstance(_dict[k], np.ndarray):
                    new_dict[k] = torch.from_numpy(_dict[k])
                else:
                    new_dict[k] = _dict[k]

                new_dict[k] = new_dict[k].to(device)
        else:
            for k in list(_dict.keys()):
                new_dict[k] = _dict[k].to(device)

        return new_dict

    def grads_to_numpy(self, g):
        if self.runs_distributed:  # if local, skip torch <--> numpy
            if g is None:
                return None
            for name in list(g.keys()):
                if g[name] is not None:
                    g[name] = g[name].cpu().numpy()
        return g

    def grads_to_torch(self, g, device):
        if self.runs_distributed:  # if local, skip torch <--> numpy
            if g is None:
                return None
            for k in list(g.keys()):
                if g[k] is not None:
                    g[k] = torch.from_numpy(g[k])

        for k in list(g.keys()):
            if g[k] is not None:
                g[k] = g[k].to(device)

        return g



# Copyright (c) 2019 Eric Steinberger


import unittest
from unittest import TestCase

import numpy as np

from PokerRL.game.Poker import Poker
from PokerRL.game.PokerRange import PokerRange
from PokerRL.game.games import DiscretizedNLHoldem, DiscretizedNLLeduc
from PokerRL.game.wrappers import VanillaEnvBuilder


class TestGeneralStaticFunctions(TestCase):

    def test_get_range_size(self):
        assert PokerRange.get_range_size(n_hole_cards=2, n_cards_in_deck=52) == 1326
        assert PokerRange.get_range_size(n_hole_cards=4, n_cards_in_deck=52) == 270725
        assert PokerRange.get_range_size(n_hole_cards=1, n_cards_in_deck=52) == 52
        assert PokerRange.get_range_size(n_hole_cards=1, n_cards_in_deck=6) == 6
        assert PokerRange.get_range_size(n_hole_cards=1, n_cards_in_deck=6) == 6
        assert PokerRange.get_range_size(n_hole_cards=3, n_cards_in_deck=6) == 20


class TestRangeInLeduc(TestCase):

    def test_create(self):
        env_bldr = get_leduc_env_bldr()
        range_ = PokerRange(env_bldr=env_bldr)

        assert np.allclose(np.sum(range_._range.reshape(-1, env_bldr.rules.RANGE_SIZE), axis=1), 1, atol=0.0001)

    def test_save_load(self):
        env_bldr = get_leduc_env_bldr()
        range_ = PokerRange(env_bldr=env_bldr)
        range_.load_state_dict(range_.state_dict())

    def test_normalize(self):
        env_bldr = get_leduc_env_bldr()
        range_ = PokerRange(env_bldr=env_bldr)

        range_._range = np.random.random(size=env_bldr.rules.RANGE_SIZE)
        range_.normalize()
        np.testing.assert_allclose(np.sum(range_._range), 1, atol=0.0001)

    def test_normalize_all_zero(self):
        env_bldr = get_leduc_env_bldr()
        range_ = PokerRange(env_bldr=env_bldr)

        range_._range = np.zeros_like(range_._range)
        range_.normalize()
        np.testing.assert_allclose(np.sum(range_._range), 1, atol=0.0001)

    def test_get_new_blockers_1d_leduc(self):
        env_bldr = get_leduc_env_bldr()
        range_ = PokerRange(env_bldr=env_bldr)

        full_board = np.array([[2, 1]], dtype=np.int8)

        should_be = {
            Poker.PREFLOP: env_bldr.lut_holder.get_1d_cards(full_board[:0]),
            Poker.FLOP: env_bldr.lut_holder.get_1d_cards(full_board),
        }

        for _round in [Poker.FLOP]:
            _n = env_bldr.lut_holder.DICT_LUT_N_CARDS_OUT[Poker.FLOP] - env_bldr.lut_holder.DICT_LUT_N_CARDS_OUT[_round]

            if _round == Poker.FLOP:
                board_2d = np.copy(full_board)
            else:
                board_2d = np.concatenate((full_board[:env_bldr.lut_holder.DICT_LUT_N_CARDS_OUT[_round]],
                                           np.array([Poker.CARD_NOT_DEALT_TOKEN_2D for _ in range(_n)],
                                                    dtype=np.int8)))

            result = range_._get_new_blockers_1d(game_round=_round, board_2d=board_2d)
            assert np.array_equal(a1=result, a2=should_be[_round])

    def test_remove_cards_from_raw_range_leduc(self):
        leduc_cards = np.array([[1, 1]], dtype=np.int8)

        env_bldr = get_leduc_env_bldr()
        range_ = PokerRange(env_bldr=env_bldr)

        range_.set_cards_to_zero_prob(cards_2d=leduc_cards)

        _ra = range_._range.reshape(-1, env_bldr.rules.RANGE_SIZE)
        for i in range(_ra.shape[0]):
            np.testing.assert_allclose(np.sum(_ra[i]), 1, atol=0.00001)

        _assert_cards_not_in_ranges(cards_2d=leduc_cards, ranges=_ra, rules=env_bldr.rules,
                                    lut_holder=env_bldr.lut_holder)

    def test_get_card_probs_leduc(self):
        env_bldr = get_leduc_env_bldr()
        range_ = PokerRange(env_bldr=env_bldr)

        assert np.array_equal(range_.get_card_probs(), range_._range)

    def test_get_possible_range_idxs_leduc(self):

        for n in range(2, 9):
            env_bldr = get_leduc_env_bldr()

            # if actually blocked
            for c in range(env_bldr.rules.N_CARDS_IN_DECK):
                board_2d = env_bldr.lut_holder.get_2d_cards(np.array([c], dtype=np.int32))
                result = PokerRange.get_possible_range_idxs(rules=env_bldr.rules, lut_holder=env_bldr.lut_holder,
                                                            board_2d=board_2d)

                should_be = np.delete(np.arange(env_bldr.rules.RANGE_SIZE, dtype=np.int32), c)

                assert np.array_equal(a1=result, a2=should_be)

            # if nothing blocked
            board_2d = np.array([Poker.CARD_NOT_DEALT_TOKEN_2D], dtype=np.int8)
            result = PokerRange.get_possible_range_idxs(rules=env_bldr.rules, lut_holder=env_bldr.lut_holder,
                                                        board_2d=board_2d)

            should_be = np.arange(env_bldr.rules.RANGE_SIZE, dtype=np.int32)

            assert np.array_equal(a1=result, a2=should_be)


class TestRangeInHoldem(TestCase):

    def test_create(self):
        env_bldr = get_holdem_env_bldr()
        range_ = PokerRange(env_bldr=env_bldr)

        assert np.allclose(np.sum(range_._range.reshape(-1, env_bldr.rules.RANGE_SIZE), axis=1), 1, atol=0.0001)

    def test_save_load(self):
        env_bldr = get_holdem_env_bldr()
        range_ = PokerRange(env_bldr=env_bldr)
        range_.load_state_dict(range_.state_dict())

    def test_get_new_blockers_1d_holdem(self):
        env_bldr = get_holdem_env_bldr()

        range_ = PokerRange(env_bldr=env_bldr)

        full_board = np.array([[1, 2], [3, 3], [12, 1], [5, 2], [6, 0]], dtype=np.int8)

        should_be = {
            Poker.PREFLOP: env_bldr.lut_holder.get_1d_cards(full_board[:0]),
            Poker.FLOP: env_bldr.lut_holder.get_1d_cards(full_board[0:3]),
            Poker.TURN: env_bldr.lut_holder.get_1d_cards(full_board[3:4]),
            Poker.RIVER: env_bldr.lut_holder.get_1d_cards(full_board[4:5]),
        }

        for _round in [Poker.PREFLOP, Poker.FLOP, Poker.TURN, Poker.RIVER]:
            _n = env_bldr.lut_holder.DICT_LUT_N_CARDS_OUT[Poker.RIVER] - \
                 env_bldr.lut_holder.DICT_LUT_N_CARDS_OUT[_round]

            if _round == Poker.RIVER:
                board_2d = np.copy(full_board)
            else:
                board_2d = np.concatenate(
                    (full_board[:env_bldr.lut_holder.DICT_LUT_N_BOARD_BRANCHES[_round]],
                     np.array([Poker.CARD_NOT_DEALT_TOKEN_2D for _ in range(_n)], dtype=np.int8))
                )

            result = range_._get_new_blockers_1d(game_round=_round, board_2d=board_2d)
            assert np.array_equal(a1=result, a2=should_be[_round])

    def test_remove_cards_from_raw_range_holdem(self):
        env_bldr = get_holdem_env_bldr()
        range_ = PokerRange(env_bldr=env_bldr)

        holdem_cards = np.array([[7, 2], [6, 0]], dtype=np.int8)

        range_.set_cards_to_zero_prob(cards_2d=holdem_cards)

        _ra = range_._range.reshape(-1, env_bldr.rules.RANGE_SIZE)
        for i in range(_ra.shape[0]):
            np.testing.assert_allclose(np.sum(_ra[i]), 1, atol=0.00001)
        _assert_cards_not_in_ranges(cards_2d=holdem_cards, ranges=_ra, rules=env_bldr.rules,
                                    lut_holder=env_bldr.lut_holder)

    def test_get_card_probs_holdem(self):
        env_bldr = get_holdem_env_bldr()
        range_ = PokerRange(env_bldr=env_bldr)

        cards_to_remove = np.array([0, 3, 6, 33, 21, 51], np.int8)

        # use previously tested method to make this test easier
        range_.set_cards_to_zero_prob(cards_2d=env_bldr.lut_holder.get_2d_cards(cards_to_remove))

        r = range_.get_card_probs()
        assert np.allclose(np.sum(r), 2, atol=0.00001)
        for c in cards_to_remove:
            assert np.allclose(r[c], 0, atol=0.00001)

    def test_get_possible_range_idxs_holdem(self):
        env_bldr = get_holdem_env_bldr()
        for n in range(2, 9):
            board_2d = np.array([[0, 0], [5, 2], [12, 3], Poker.CARD_NOT_DEALT_TOKEN_2D], dtype=np.int8)
            result = PokerRange.get_possible_range_idxs(rules=env_bldr.rules, lut_holder=env_bldr.lut_holder,
                                                        board_2d=board_2d)

            assert result.shape[0] == 1176  # nck of 49:2

            # all of these should be blocked
            for e in [0, 1, 2, 3, 4, 50, 1325]:
                assert not np.any(result == e)


def _assert_cards_not_in_ranges(cards_2d, ranges, rules, lut_holder):
    if rules.N_HOLE_CARDS == 1:  # leduc
        for p in range(ranges.shape[0]):
            for c in cards_2d:
                np.testing.assert_allclose(
                    ranges[p, lut_holder.get_range_idx_from_hole_cards(hole_cards_2d=c.reshape(-1, 2))], 0,
                    atol=0.0001)

    elif rules.N_HOLE_CARDS == 2:  # holdem
        cards_1d = lut_holder.get_1d_cards(cards_2d=cards_2d.reshape(-1, 2))
        for p in range(ranges.shape[0]):
            for c in cards_1d:
                for c1 in range(0, c):
                    np.testing.assert_allclose(ranges[p, lut_holder.LUT_HOLE_CARDS_2_IDX[c1, c]], 0,
                                               atol=0.00001)
                for c2 in range(c + 1, rules.N_CARDS_IN_DECK):
                    np.testing.assert_allclose(ranges[p, lut_holder.LUT_HOLE_CARDS_2_IDX[c, c2]], 0,
                                               atol=0.00001)
    else:
        raise NotImplementedError


def get_holdem_env_bldr():
    return VanillaEnvBuilder(env_cls=DiscretizedNLHoldem,
                             env_args=DiscretizedNLHoldem.ARGS_CLS(n_seats=3,
                                                                   starting_stack_sizes_list=[234] * 3,
                                                                   bet_sizes_list_as_frac_of_pot=[0.1, 0.4, 1.0],
                                                                   stack_randomization_range=(-123, 234)))


def get_leduc_env_bldr():
    return VanillaEnvBuilder(env_cls=DiscretizedNLLeduc,
                             env_args=DiscretizedNLLeduc.ARGS_CLS(n_seats=3,
                                                                  starting_stack_sizes_list=[234] * 3,
                                                                  bet_sizes_list_as_frac_of_pot=[0.1, 0.4, 1.0],
                                                                  stack_randomization_range=(-123, 234)))


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Eric Steinberger


import copy

import numpy as np

from PokerRL.game.PokerEnvStateDictEnums import EnvDictIdxs
from PokerRL.game._.wrappers._Wrapper import Wrapper


class RecurrentHistoryWrapper(Wrapper):
    """
    Stores a sequence of current observations of each timestep thus having perfect recall.
    This wrapper is suitable for recurrent NN architectures.
    """

    def __init__(self, env, env_bldr_that_built_me):
        super().__init__(env=env, env_bldr_that_built_me=env_bldr_that_built_me)
        self.invert_history_order = env_bldr_that_built_me.invert_history_order
        self._list_of_obs_this_episode = None

    def _reset_state(self, **kwargs):
        self._list_of_obs_this_episode = []

    def _pushback(self, env_obs):
        if self.invert_history_order:
            self._list_of_obs_this_episode.insert(0, np.copy(env_obs))
        else:
            self._list_of_obs_this_episode.append(np.copy(env_obs))

    def print_obs(self, wrapped_obs):
        assert isinstance(wrapped_obs, np.ndarray)
        print("*****************************************************************************************************")
        print("*****************************************************************************************************")
        print("*****************************************************************************************************")
        print()
        print("________________________________________ OBSERVATION HISTORY ________________________________________")
        print()

        for o in wrapped_obs:
            self.env.print_obs(o)

    def get_current_obs(self, env_obs=None):
        return np.array(self._list_of_obs_this_episode, dtype=np.float32)

    def state_dict(self):
        return {
            "base": super().state_dict(),
            "obs_seq": copy.deepcopy(self._list_of_obs_this_episode)
        }

    def load_state_dict(self, state_dict):
        super().load_state_dict(state_dict=state_dict["base"])
        self._list_of_obs_this_episode = copy.deepcopy(state_dict["obs_seq"])

    def set_to_public_tree_node_state(self, node):
        """
        Sets the internal env wrapper to the state ""node"" is in.

        Args:
            node:                   Any node (of any type) in a PublicTree instance.

        """
        state_seq = []  # will be sorted. [0] is root.

        def add(_node):
            if _node is not None:
                if _node.p_id_acting_next != _node.tree.CHANCE_ID:
                    self.env.load_state_dict(_node.env_state)
                    state_seq.insert(0, self.env.get_current_obs(is_terminal=False))
                add(_node.parent)

        add(node)  # will step from node to parent to parent of parent... to root.
        self.reset()
        self._reset_state()  # from .reset() the first obs is in by default

        # fake step in the internal env to gain current rr state
        for obs in state_seq:
            self._pushback(env_obs=obs)

        assert len(state_seq) == len(self._list_of_obs_this_episode)

        self.env.load_state_dict(node.env_state, blank_private_info=True)
        assert np.array_equal(node.env_state[EnvDictIdxs.board_2d], self.env.board)



# Copyright (c) 2019 Eric Steinberger
# Inspiration of architecture from DeepStack-Leduc (https://github.com/lifrordi/DeepStack-Leduc/tree/master/Source)

import copy
import json
import os

import numpy as np

from PokerRL.game.Poker import Poker
from PokerRL.game.PokerEnvStateDictEnums import EnvDictIdxs, PlayerDictIdxs
from PokerRL.game._.tree._.StrategyFiller import StrategyFiller
from PokerRL.game._.tree._.ValueFiller import ValueFiller
from PokerRL.game._.tree._.nodes import PlayerActionNode, ChanceNode
from PokerRL.util import file_util


class PublicTree:
    """
    Builds a full public game tree to compute the counterfactual values and the best response against strategy profiles.
    You can also visualize game trees in the browser as described in README.md.

    Warning: This part of PokerRL is INEFFICIENT and SLOW! Only works in Leduc games (no limit Leduc too though)
    The reason why we don't really care though is that there are great codebases for CFR and alike in large games.
    Deep Learning methods spend most of the time evaluating neural nets for strategy queries. This is no different even
    with this horribly slow implementation of a Public Tree; its just that the tree has 5% instead of 0.001% overhead.
    """
    CHANCE_ID = "Ch"

    def __init__(self,
                 env_bldr,
                 stack_size,
                 stop_at_street,
                 put_out_new_round_after_limit=False,
                 is_debugging=False,
                 ):
        """
        To start the tree from a given scenario, set ""env"" to that scenario and it will be treated
        as the root.
        """
        self._env_bldr = env_bldr
        self._is_debugging = is_debugging
        self._stack_size = stack_size

        # set up optional graphical tree export
        _dir_tree_vis = os.path.join("C:\\" if os.name == 'nt' else os.path.expanduser('~/'), "PokerRL_Viz")

        # This is just me being over-cautious to not overwrite someones files...
        _tree_vis_installed = os.path.isdir(_dir_tree_vis) \
                              and os.path.isfile(os.path.join(_dir_tree_vis, "ALLOWED_TO_WRITE_HERE.dontdelete"))

        if _tree_vis_installed:
            self.dir_tree_vis_data = os.path.join(_dir_tree_vis, "data")
            file_util.create_dir_if_not_exist(self.dir_tree_vis_data)
        else:
            self.dir_tree_vis_data = None

        # _________________________________ tree vars __________________________________
        self.root = None
        self._n_nodes = 0
        self._n_nonterm = 0

        self._env = self._env_bldr.get_new_env(is_evaluating=True, stack_size=stack_size)
        _ar = self._env.get_args()
        _ar.RETURN_PRE_TRANSITION_STATE_IN_INFO = True
        self._env.set_args(_ar)
        self._env.reset()

        self._n_seats = self._env_bldr.N_SEATS

        # _____________________ stops building tree at given round _____________________
        self._stop_at_street = max(self._env.ALL_ROUNDS_LIST) + 1 if stop_at_street is None else stop_at_street
        self._put_out_new_round_after_limit = put_out_new_round_after_limit

        self._value_filler = ValueFiller(tree=self)
        self._strategy_filler = StrategyFiller(tree=self, env_bldr=env_bldr)

    # _____________________________________________________ API ________________________________________________________
    @property
    def stack_size(self):
        return self._stack_size

    @property
    def is_debugging(self):
        return self._is_debugging

    @property
    def n_nodes(self):
        return self._n_nodes

    @property
    def n_nonterm(self):
        return self._n_nonterm

    @property
    def n_seats(self):
        return self._n_seats

    @property
    def stop_at_street(self):
        return self._stop_at_street

    @property
    def put_out_new_round_after_limit(self):
        return self._put_out_new_round_after_limit

    @property
    def env_bldr(self):
        return self._env_bldr

    def build_tree(self):
        """
        Builds from the current state of the environment
        """
        self.root = ChanceNode(env_state=self._env.state_dict(),
                               tree=self,
                               parent=None,
                               is_terminal=False,
                               p_id_acted_last=None,
                               depth=0)
        self.root.allowed_actions = self._env.get_legal_actions()
        self.root.reach_probs = np.full(shape=(self._env_bldr.N_SEATS, self._env_bldr.rules.RANGE_SIZE),
                                        fill_value=1.0 / float(self._env_bldr.rules.RANGE_SIZE),
                                        dtype=np.float32)

        self._build_tree(current_node=self.root)

    def compute_ev(self):
        self._value_filler.compute_cf_values_heads_up(self.root)

    def fill_uniform_random(self):
        self._strategy_filler.fill_uniform_random()

    def fill_random_random(self):
        self._strategy_filler.fill_random_random()

    def fill_with_agent_policy(self, agent):
        self._strategy_filler.fill_with_agent_policy(agent=agent)

    def update_reach_probs(self):
        self._strategy_filler.update_reach_probs()

    def get_tree_as_dict(self):
        return self._export_for_node_strategy_tree(self.root)

    def export_to_file(self, name="data"):
        if self.dir_tree_vis_data is not None:
            file_util.write_dict_to_file_js(_dir=self.dir_tree_vis_data, file_name=name,
                                            dictionary=self.get_tree_as_dict())

    def copy(self):
        _tree = PublicTree(env_bldr=self._env_bldr,
                           stack_size=copy.deepcopy(self._stack_size),
                           stop_at_street=self._stop_at_street,
                           put_out_new_round_after_limit=self._put_out_new_round_after_limit)
        _tree.root = copy.deepcopy(self.root)

        return _tree

    # __________________________________________________ INTERNAL ______________________________________________________
    def _build_tree(self, current_node):
        current_node.children = self._get_children_nodes(node=current_node)
        self._n_nodes += len(current_node.children)
        self._n_nonterm += len([c for c in current_node.children if (not c.is_terminal)])
        for child in current_node.children:
            self._build_tree(current_node=child)

    def _get_children_of_action_node(self, parent):
        """
        Creates the children nodes after an action node (a node where a player acted).
        """
        self._env.load_state_dict(parent.env_state)
        if self._stop_at_street <= self._env.current_round:
            return []

        if parent.p_id_acting_next == self.CHANCE_ID:
            """
            When we get here, the parent node has a special attribute that carries the state after the new round is
            started. Here we adjust that state to not have the newly dealt cards dealt and cards drawn from the deck,
            but all other updates stay.
            """

            # ______________________________________ break if too far ______________________________________
            if not self._put_out_new_round_after_limit:
                if self._stop_at_street <= parent.env_state[EnvDictIdxs.current_round]:
                    return []

            # _____________________ prepare state to new round without new cards dealt _____________________
            _base_env_state = copy.deepcopy(parent.new_round_state)

            # ________________________________ generate all possible boards ________________________________

            def _make_boards(_new_boards, _board_1d, _idx_to_insert_in, _n_todo):
                if _n_todo > 0:
                    for _c in [_ for _ in range(self._env.N_CARDS_IN_DECK) if _ not in _board_1d]:
                        _b = np.copy(_board_1d)
                        _b[_idx_to_insert_in] = _c

                        for i in range(_n_todo):
                            _make_boards(_new_boards=_new_boards, _board_1d=_b, _idx_to_insert_in=_idx_to_insert_in + 1,
                                         _n_todo=_n_todo - 1)

                        _new_boards.append(self._env_bldr.lut_holder.get_2d_cards(cards_1d=_b))

            new_boards = []
            board_1d = self._env_bldr.lut_holder.get_1d_cards(cards_2d=_base_env_state[EnvDictIdxs.board_2d])
            _make_boards(_new_boards=new_boards,
                         _board_1d=board_1d,
                         _idx_to_insert_in=self._env_bldr.lut_holder.DICT_LUT_N_CARDS_OUT[
                             self._env_bldr.rules.ROUND_BEFORE[_base_env_state[EnvDictIdxs.current_round]]],
                         _n_todo=self._env_bldr.lut_holder.DICT_LUT_CARDS_DEALT_IN_TRANSITION_TO[
                             _base_env_state[EnvDictIdxs.current_round]])

            # ____________________________________ build children nodes ____________________________________
            children = []
            for board in new_boards:
                child_env_state = copy.deepcopy(_base_env_state)
                child_env_state[EnvDictIdxs.board_2d] = board
                _deck_remaining_before_deal = np.copy(child_env_state[EnvDictIdxs.deck]["deck_remaining"])
                for c in board:
                    child_env_state[EnvDictIdxs.deck]["deck_remaining"] = \
                        np.delete(_deck_remaining_before_deal, np.where(_deck_remaining_before_deal == c), axis=0)

                c = ChanceNode(env_state=child_env_state,
                               tree=self,
                               parent=parent,
                               is_terminal=False,
                               p_id_acted_last=self.CHANCE_ID,
                               depth=parent.depth + 1)

                self._env.load_state_dict(child_env_state)
                c.allowed_actions = self._env.get_legal_actions()
                children.append(c)

            return children

        else:
            children = []
            for action in parent.allowed_actions:
                self._env.load_state_dict(parent.env_state)
                _, __, is_terminal, info = self._env.step(action)

                # after action:  Terminal
                if is_terminal:
                    # state before payouts or rundown
                    env_state = info["state_dict_before_money_move"]
                    env_state[EnvDictIdxs.current_round] = parent.env_state[EnvDictIdxs.current_round]
                    env_state[EnvDictIdxs.board_2d] = parent.env_state[EnvDictIdxs.board_2d]
                    env_state[EnvDictIdxs.deck] = copy.deepcopy(parent.env_state[EnvDictIdxs.deck])
                    is_terminal = True
                    new_round_state = None

                # after action:  New round
                elif info["chance_acts"]:
                    # state before round transition
                    env_state = info["state_dict_before_money_move"]
                    env_state[EnvDictIdxs.current_round] = parent.env_state[EnvDictIdxs.current_round]
                    if self._is_debugging:
                        assert env_state[EnvDictIdxs.current_round] == parent.env_state[EnvDictIdxs.current_round]
                        assert np.array_equal(env_state[EnvDictIdxs.board_2d],
                                              np.copy(parent.env_state[EnvDictIdxs.board_2d]))
                        assert np.array_equal(env_state[EnvDictIdxs.deck]["deck_remaining"],
                                              np.copy(parent.env_state[EnvDictIdxs.deck]["deck_remaining"]))
                    is_terminal = False
                    new_round_state = self._env.state_dict()
                    new_round_state[EnvDictIdxs.board_2d] = np.copy(parent.env_state[EnvDictIdxs.board_2d])
                    new_round_state[EnvDictIdxs.deck] = copy.deepcopy(parent.env_state[EnvDictIdxs.deck])

                else:
                    env_state = self._env.state_dict()
                    is_terminal = False
                    new_round_state = None

                node = PlayerActionNode(env_state=env_state,
                                        action=action,
                                        is_terminal=is_terminal,
                                        parent=parent,
                                        p_id_acted_last=parent.p_id_acting_next,
                                        tree=self,
                                        depth=parent.depth + 1,
                                        new_round_state=new_round_state)

                if is_terminal:
                    node.p_id_acting_next = None
                    node.allowed_actions = []
                elif info["chance_acts"]:
                    node.p_id_acting_next = self.CHANCE_ID
                    node.allowed_actions = []
                else:
                    node.allowed_actions = self._env.get_legal_actions()
                children.append(node)

            return children

    def _get_children_nodes(self, node):
        if node.is_terminal:
            return []
        else:
            return self._get_children_of_action_node(parent=node)  # -> action or chance or terminal

    def _get_action_as_str(self, action):
        if action is None:
            return "None"
        mapping = {
            Poker.FOLD: "FOLD",
            Poker.CHECK_CALL: "CHECK",
            "CHANCE": "CHANCE"
        }

        # Returns RX where X is the index of the bet-size option selected, if the player raises/bets
        return mapping.get(action, "R" + str(action - 2))

    def _export_for_node_strategy_tree(self, node):

        def _2darr_to_str(arr):
            _str = ""
            if arr is None:
                return "Not Computed"

            for i in range(arr.shape[0]):
                for j in range(arr.shape[1]):
                    _str += str("{:10.4f}".format(arr[i, j])) + " "
                if i < arr.shape[0] - 1:
                    _str += " || "

            return _str

        if node.parent is None:
            title = "ROOT"
        elif isinstance(node, ChanceNode):
            title = self._env.cards2str(cards_2d=node.env_state[EnvDictIdxs.board_2d])
        else:
            title = "Player acted last " + str(node.p_id_acted_last) + " :: Action: " + self._get_action_as_str(
                node.action) + " :: Board: " + self._env.cards2str(cards_2d=node.env_state[EnvDictIdxs.board_2d])

        """
        Tree print format for this field is (note, in each state only one player acts): 
        pi_p(a1, h1) pi_p(a2, h1) ... pi_p(an, h1) ... || pi_p(a1, h2) pi_p(a2, h2) ... pi_p(an, h2) || ...
        """
        strat_str = _2darr_to_str(node.strategy)

        """
        Tree print format for this field is: 
        reach(p1, h1) reach(p1, h2) ... reach(p1, hn) || reach(p2, h1) reach(p2, h2) ... reach(p2, hn)
        """
        reach_probs_str = _2darr_to_str(node.reach_probs)

        """
        Tree print format for this field is: 
        ev(p1, h1) ev(p1, h2) ... ev(p1, hn) || ev(p2, h1) ev(p2, h2) ... ev(p2, hn)
        """
        ev_str = _2darr_to_str(node.ev)

        """
        Tree print format for this field is: 
        ev_br(p1, h1) ev_br(p1, h2) ... ev_br(p1, hn) || ev_br(p2, h1) ev_br(p2, h2) ... ev_br(p2, hn)
        """
        ev_br_str = _2darr_to_str(node.ev_br)

        if isinstance(node, np.ndarray) and len(node.shape) == 2:
            """ Since the data field is general, it is impossible to know how/what to format. The code below is capable
                of processing 2d numpy arrays. For anything else, we will set an empty string. """
            data_str = _2darr_to_str(node.data)
        else:
            data_str = ""

        json_table = {
            'text': {
                # for header
                'title': title,
                'round': "Round : " + Poker.INT2STRING_ROUND[node.env_state[EnvDictIdxs.current_round]],

                # for body
                'main_pot': "Pot : " + json.dumps(int(node.env_state[EnvDictIdxs.main_pot])),

                'terminal': ("TERM" if str(node.is_terminal) else "") + " " + str(node.allowed_actions),

                'side_pots': "SP: " + json.dumps(
                    [int(i) for i in node.env_state[EnvDictIdxs.side_pots]]),

                'stack_sizes': "Stacks: " + json.dumps(
                    [int(s[PlayerDictIdxs.stack]) for s in node.env_state[EnvDictIdxs.seats]]),

                'current_bets': "Bets: " + json.dumps(
                    [int(s[PlayerDictIdxs.current_bet]) for s in node.env_state[EnvDictIdxs.seats]]),

                'not_folded': "Playing: " + json.dumps(
                    [not s[PlayerDictIdxs.folded_this_episode] for s in node.env_state[EnvDictIdxs.seats]]).replace(
                    "true", "1").replace("false", "0") + "  Next: " + str(node.p_id_acting_next),

                "exploitability": "Exploitability: " + (str(
                    node.exploitability) if node.exploitability is not None else "Exploitability not computed")
                                  + "   ||   BR Action per hand "
                                  + ("" if node.br_a_idx_in_child_arr_for_each_hand is None else
                                     str([self._get_action_as_str(a)
                                          for a in
                                          np.array(node.allowed_actions)[node.br_a_idx_in_child_arr_for_each_hand]])),

                'strategy': "STRAT: " + strat_str,

                'reach_probs': "REACH: " + reach_probs_str,

                'ev': "EV: " + ev_str,

                'ev_br': "EV-BR: " + ev_br_str,

                'data': "DATA: " + data_str
            },

            # tree nodes are all collapsed by default
            'collapsed': True,

            'children': []
        }

        # recurse over all children
        for child in node.children:
            json_table['children'].append(self._export_for_node_strategy_tree(child))

        return json_table



from PokerRL.game._.rl_env.poker_types.DiscretizedPokerEnv import DiscretizedPokerEnv
from PokerRL.game._.rl_env.poker_types.LimitPokerEnv import LimitPokerEnv
from PokerRL.rl.rl_util import get_builder_from_str, get_env_cls_from_str


def get_env_builder_rlbr(t_prof):
    env_bldr_cls = get_builder_from_str(t_prof.env_builder_cls_str)
    return env_bldr_cls(env_cls=get_env_cls_from_str(t_prof.game_cls_str),
                        env_args=t_prof.module_args["rlbr"].get_rlbr_env_args(
                            agents_env_args=t_prof.module_args["env"]))


def reset_episode_multi_action_space(rlbr_env_wrapper, opponent_agent):
    ret = rlbr_env_wrapper.reset()
    opponent_agent.reset(deck_state_dict=rlbr_env_wrapper.env.cards_state_dict())
    return ret


def notify_agent_multi_action_space(action_int, rlbr_seat_id, rlbr_env_wrapper, opponent_agent):
    _type = type(rlbr_env_wrapper.env)

    if issubclass(_type, LimitPokerEnv):
        opponent_agent.notify_of_action(p_id_acted=rlbr_seat_id, action_he_did=action_int)

    elif issubclass(_type, DiscretizedPokerEnv):
        if action_int >= 2:
            raise_frac = rlbr_env_wrapper.env.bet_sizes_list_as_frac_of_pot[action_int - 2]
            opponent_agent.notify_of_raise_frac_action(p_id_acted=rlbr_seat_id, frac=raise_frac)
        else:
            opponent_agent.notify_of_action(p_id_acted=rlbr_seat_id, action_he_did=action_int)

    else:
        raise ValueError(_type)


def step_from_opp_action(action_int, opponent, rlbr_env_wrapper):
    _type = type(rlbr_env_wrapper.env)

    if issubclass(_type, LimitPokerEnv):
        return rlbr_env_wrapper.step(action=action_int)

    elif issubclass(_type, DiscretizedPokerEnv):
        if action_int >= 2:
            raise_frac = opponent.env_bldr.env_args.bet_sizes_list_as_frac_of_pot[action_int - 2]
            return rlbr_env_wrapper.step_raise_pot_frac(pot_frac=raise_frac)
        else:
            return rlbr_env_wrapper.step(action=action_int)

    else:
        raise ValueError(_type)



# Copyright (c) 2019 Eric Steinberger


import unittest
from unittest import TestCase

import numpy as np

from PokerRL.game._.cpp_wrappers.CppHandeval import CppHandeval


class TestCppLib(TestCase):

    def test_get_hand_rank_52_holdem(self):
        cpp_poker = CppHandeval()
        b = np.array([[2, 0], [2, 3], [11, 1], [10, 2], [11, 2]], dtype=np.int8)
        h = np.array([[11, 3], [5, 1]], dtype=np.int8)
        assert isinstance(cpp_poker.get_hand_rank_52_holdem(hand_2d=h, board_2d=b), int)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Eric Steinberger


import os
from os.path import join as ospj

import numpy as np

from PokerRL._.CppWrapper import CppWrapper
from PokerRL.game.Poker import Poker
from PokerRL.game._.rl_env.game_rules import HoldemRules


class CppLibHoldemLuts(CppWrapper):

    def __init__(self, n_boards_lut, n_cards_out_lut):
        super().__init__(path_to_dll=ospj(os.path.dirname(os.path.realpath(__file__)),
                                          "lib_luts." + self.CPP_LIB_FILE_ENDING))
        self._n_boards_lut = n_boards_lut
        self._n_cards_out_lut = n_cards_out_lut

        self._clib.get_hole_card_2_idx_lut.argtypes = [self.ARR_2D_ARG_TYPE]
        self._clib.get_hole_card_2_idx_lut.restype = None

        self._clib.get_idx_2_hole_card_lut.argtypes = [self.ARR_2D_ARG_TYPE]
        self._clib.get_idx_2_hole_card_lut.restype = None

        self._clib.get_idx_2_flop_lut.argtypes = [self.ARR_2D_ARG_TYPE]
        self._clib.get_idx_2_flop_lut.restype = None

        self._clib.get_idx_2_turn_lut.argtypes = [self.ARR_2D_ARG_TYPE]
        self._clib.get_idx_2_turn_lut.restype = None

        self._clib.get_idx_2_river_lut.argtypes = [self.ARR_2D_ARG_TYPE]
        self._clib.get_idx_2_river_lut.restype = None

    # __________________________________________________ LUTs __________________________________________________________
    def get_idx_2_hole_card_lut(self):
        lut = np.full(shape=(HoldemRules.RANGE_SIZE, 2), fill_value=-2, dtype=np.int8)
        self._clib.get_idx_2_hole_card_lut(self.np_2d_arr_to_c(lut))  # fills it
        return lut

    def get_hole_card_2_idx_lut(self):
        lut = np.full(shape=(HoldemRules.N_CARDS_IN_DECK, HoldemRules.N_CARDS_IN_DECK),
                      fill_value=-2, dtype=np.int16)
        self._clib.get_hole_card_2_idx_lut(self.np_2d_arr_to_c(lut))  # fills it
        return lut

    def get_idx_2_flop_lut(self):
        lut = np.full(shape=(
            self._n_boards_lut[Poker.FLOP],
            self._n_cards_out_lut[Poker.FLOP]),
            fill_value=-2, dtype=np.int8)
        self._clib.get_idx_2_flop_lut(self.np_2d_arr_to_c(lut))  # fills it
        return lut

    def get_idx_2_turn_lut(self):
        lut = np.full(shape=(
            self._n_boards_lut[Poker.TURN],
            self._n_cards_out_lut[Poker.TURN]),
            fill_value=-2, dtype=np.int8)
        self._clib.get_idx_2_turn_lut(self.np_2d_arr_to_c(lut))  # fills it
        return lut

    def get_idx_2_river_lut(self):
        lut = np.full(shape=(
            self._n_boards_lut[Poker.RIVER],
            self._n_cards_out_lut[Poker.RIVER]),
            fill_value=-2, dtype=np.int8)
        self._clib.get_idx_2_river_lut(self.np_2d_arr_to_c(lut))  # fills it
        return lut

    def get_1d_card(self, card_2d):
        """
        Args:
            card_2d (np.ndarray):    array of 2 int8s. [rank, suit]

        Returns:
            int8: 1d representation of card_2d

        """
        return self._clib.get_1d_card(self.np_1d_arr_to_c(card_2d))

    def get_2d_card(self, card_1d):
        """
        Args:
            card_1d (int): 

        Returns:
            np.ndarray(shape=2, dtype=np.int8): 2d representation of card_1d
        """
        card_2d = np.empty(shape=2, dtype=np.int8)
        self._clib.get_2d_card(card_1d, self.np_1d_arr_to_c(card_2d))
        return card_2d



# Copyright (c) 2019 Eric Steinberger


import numpy as np


class DeckOfCards:
    """
    Cards are stored in 2D form [rank, suit]
    """

    def __init__(self, num_suits=4, num_ranks=13):
        self.n_suits = num_suits
        self.n_ranks = num_ranks

        self.deck_remaining = self._build_deck()
        self._starting_deck = self._build_deck()
        self.shuffle()

    def shuffle(self):
        np.random.shuffle(self.deck_remaining)

    def draw(self, num_cards):
        """ draws from top. is completely deterministic from a given starting state """
        cards = self.deck_remaining[:num_cards]
        self.deck_remaining = self.deck_remaining[num_cards:]
        return cards

    def reset(self):
        self.deck_remaining = np.copy(self._starting_deck)
        self.shuffle()

    def state_dict(self):
        return {
            "deck_remaining": np.copy(self.deck_remaining)
        }

    def load_state_dict(self, state_dict):
        self.deck_remaining = np.copy(state_dict["deck_remaining"])

    def _build_deck(self):
        deck = np.empty(shape=(self.n_ranks * self.n_suits, 2), dtype=np.int8)

        for r in range(self.n_ranks):
            i = self.n_suits * r
            deck[i:i + self.n_suits, 0] = r
            for s in range(self.n_suits):
                deck[i + s, 1] = s

        return deck

    def remove_cards(self, cards_2d):
        idxs_to_del = []
        for to_rem in cards_2d:
            for i, c in enumerate(self.deck_remaining):
                if np.array_equal(c, to_rem):
                    idxs_to_del.append(i)
                    break

        self.deck_remaining = np.delete(self.deck_remaining, idxs_to_del, axis=0)



# Copyright (c) 2019 Eric Steinberger


import os

import psutil

from PokerRL.eval.rl_br import _util
from PokerRL.rl import rl_util
from PokerRL.rl.agent_modules.DDQN import DDQN
from PokerRL.rl.base_cls.workers.ParameterServerBase import ParameterServerBase as _ParameterServerBase
from PokerRL.rl.neural.DuelingQNet import DuelingQNet


class Local_RLBR_ParameterServer(_ParameterServerBase):

    def __init__(self, t_prof, chief_handle):
        super().__init__(t_prof=t_prof, chief_handle=chief_handle)
        self._args = t_prof.module_args["rlbr"]
        self._eval_env_bldr = _util.get_env_builder_rlbr(t_prof=t_prof)
        self._env_bldr = self._eval_env_bldr  # Override base class variable

        if self._t_prof.log_verbose:
            self._exp_mem_usage = self._ray.get(
                self._ray.remote(self._chief_handle.create_experiment,
                                 self._t_prof.name + "_RL-BR_PS_Memory_Usage")
            )

        self._nets = [None for _ in range(self._env_bldr.N_SEATS)]
        self._eps = [None for _ in range(self._env_bldr.N_SEATS)]
        self._optims = [None for _ in range(self._env_bldr.N_SEATS)]

    def reset(self, p_id, global_iteration):
        self._nets[p_id] = self._get_new_net()
        self._optims[p_id] = self._get_new_optim(p_id=p_id)
        self.update_eps(p_id=p_id, update_nr=0)

        if self._t_prof.log_verbose:
            # Logs
            process = psutil.Process(os.getpid())
            self._ray.remote(self._chief_handle.add_scalar,
                             self._exp_mem_usage, "Debug/MemoryUsage/PS", global_iteration, process.memory_info().rss)

    def get_weights(self, p_id):
        self._nets[p_id].zero_grad()
        return self._ray.state_dict_to_numpy(self._nets[p_id].state_dict())

    def apply_grads(self, p_id, list_of_grads):
        self._apply_grads(list_of_grads=list_of_grads, optimizer=self._optims[p_id], net=self._nets[p_id],
                          grad_norm_clip=self._args.ddqn_args.grad_norm_clipping)

    def update_eps(self, p_id, update_nr):
        self._eps[p_id] = rl_util.polynomial_decay(base=self._args.ddqn_args.eps_start,
                                                   minimum=self._args.ddqn_args.eps_min,
                                                   const=self._args.ddqn_args.eps_const,
                                                   exponent=self._args.ddqn_args.eps_exponent,
                                                   counter=update_nr)

    def get_eps(self, p_id):
        return self._eps[p_id]

    def _get_new_net(self):
        return DuelingQNet(q_args=self._args.ddqn_args.q_args, env_bldr=self._eval_env_bldr, device=self._device)

    def _get_new_optim(self, p_id):
        return rl_util.str_to_optim_cls(self._args.ddqn_args.optim_str)(self._nets[p_id].parameters(),
                                                                        lr=self._args.ddqn_args.lr)

    def get_eval_ddqn_state_dicts(self):
        ddqns = []
        for p in range(self._eval_env_bldr.N_SEATS):
            ddqn = DDQN(owner=p, ddqn_args=self._args.ddqn_args, env_bldr=self._eval_env_bldr)
            ddqn.load_net_state_dict(self._nets[p].state_dict())
            ddqn.update_target_net()
            ddqn.eps = None
            ddqn.buf = None
            ddqns.append(ddqn.state_dict())
        return ddqns



# Copyright (c) 2019 Eric Steinberger


import ctypes
import os
from os.path import join as ospj

import numpy as np

from PokerRL._.CppWrapper import CppWrapper
from PokerRL.game._.rl_env.game_rules import HoldemRules


class CppHandeval(CppWrapper):

    def __init__(self):
        super().__init__(path_to_dll=ospj(os.path.dirname(os.path.realpath(__file__)),
                                          "lib_hand_eval." + self.CPP_LIB_FILE_ENDING))
        self._clib.get_hand_rank_52_holdem.argtypes = [
            self.ARR_2D_ARG_TYPE,
            self.ARR_2D_ARG_TYPE
        ]
        self._clib.get_hand_rank_52_holdem.restype = ctypes.c_int32

        self._clib.get_hand_rank_all_hands_on_given_boards_52_holdem.argtypes = [
            self.ARR_2D_ARG_TYPE,
            self.ARR_2D_ARG_TYPE,
            ctypes.c_int32,
            self.ARR_2D_ARG_TYPE,
            self.ARR_2D_ARG_TYPE
        ]
        self._clib.get_hand_rank_all_hands_on_given_boards_52_holdem.restype = None

    def get_hand_rank_52_holdem(self, hand_2d, board_2d):
        """
        Args:
            hand_2d (np.ndarray(shape=[5,2], dtype=int8)):      [rank, suit], [rank, suit]]
            board_2d (np.ndarray(shape=[5,2], dtype=int8)):     [rank, suit], [rank, suit], ...]

        Returns:
            int: integer representing strength of the strongest 5card hand in the 7 cards. higher is better.
        """
        return self._clib.get_hand_rank_52_holdem(self.np_2d_arr_to_c(hand_2d), self.np_2d_arr_to_c(board_2d))

    def get_hand_rank_all_hands_on_given_boards_52_holdem(self, boards_1d, lut_holder):
        """
        Args:
            boards_1d (np.ndarray(shape=[N, 5], dtype=int8)):   [[c1, c2, c3, c4, c5], [c1, c2, .., c5], ...}

        Returns:
            np.ndarray(shape=[N, RANGE_SIZE], dtype=int32):     hand_rank for each possible hand; -1 for
                                                                blocked on each of the given boards

        """
        assert len(boards_1d.shape) == 2
        assert boards_1d.shape[1] == 5
        hand_ranks = np.full(shape=(boards_1d.shape[0], HoldemRules.RANGE_SIZE), fill_value=-1, dtype=np.int32)
        self._clib.get_hand_rank_all_hands_on_given_boards_52_holdem(
            self.np_2d_arr_to_c(hand_ranks),  # int32**
            self.np_2d_arr_to_c(boards_1d),  # int8**
            boards_1d.shape[0],  # int (number of boards)
            self.np_2d_arr_to_c(lut_holder.LUT_IDX_2_HOLE_CARDS),  # int8**
            self.np_2d_arr_to_c(lut_holder.LUT_1DCARD_2_2DCARD)  # int8**
        )
        return hand_ranks



name = "PokerRL"

from . import cfr
from . import eval
from . import game
from . import rl
from . import util

from .game.games import *
from .game.Poker import *
from .game.AgentTournament import *
from .game.InteractiveGame import *
from .game.wrappers import *



# Copyright (c) 2019 Eric Steinberger


import numpy as np
import torch

from PokerRL.rl import rl_util
from PokerRL.rl.neural.DuelingQNet import DuelingQNet
from PokerRL.rl.neural.NetWrapperBase import NetWrapperArgsBase as _NetWrapperArgsBase
from PokerRL.rl.neural.NetWrapperBase import NetWrapperBase as _NetWrapperBase


class DDQN(_NetWrapperBase):

    def __init__(self,
                 env_bldr,
                 ddqn_args,
                 owner,
                 ):
        super().__init__(
            net=DuelingQNet(env_bldr=env_bldr, q_args=ddqn_args.q_args, device=ddqn_args.device_training),
            env_bldr=env_bldr,
            args=ddqn_args,
            owner=owner,
            device=ddqn_args.device_training,
        )

        self._eps = None

        self._target_net = DuelingQNet(env_bldr=env_bldr, q_args=ddqn_args.q_args, device=ddqn_args.device_training)
        self._target_net.eval()
        self.update_target_net()

        self._batch_arranged = torch.arange(ddqn_args.batch_size, dtype=torch.long, device=self.device)
        self._minus_e20 = torch.full((ddqn_args.batch_size, self._env_bldr.N_ACTIONS,),
                                     fill_value=-10e20,
                                     device=self.device,
                                     dtype=torch.float32,
                                     requires_grad=False)

        self._n_actions_arranged = np.arange(self._env_bldr.N_ACTIONS, dtype=np.int32).tolist()

    @property
    def eps(self):
        return self._eps

    @eps.setter
    def eps(self, value):
        self._eps = value

    def select_br_a(self, pub_obses, range_idxs, legal_actions_lists, explore=False):
        if explore and (np.random.random() < self._eps):
            return np.array(
                [legal_actions[np.random.randint(len(legal_actions))] for legal_actions in legal_actions_lists]
            )

        with torch.no_grad():
            self.eval()
            range_idxs = torch.tensor(range_idxs, dtype=torch.long, device=self.device)
            q = self._net(pub_obses=pub_obses, range_idxs=range_idxs,
                          legal_action_masks=rl_util.batch_get_legal_action_mask_torch(
                              n_actions=self._env_bldr.N_ACTIONS,
                              legal_actions_lists=legal_actions_lists,
                              device=self.device,
                              dtype=torch.float32)).cpu().numpy()
            for b in range(q.shape[0]):
                illegal_actions = [i for i in self._n_actions_arranged if i not in legal_actions_lists[b]]
                if len(illegal_actions) > 0:
                    illegal_actions = np.array(illegal_actions)
                    q[b, illegal_actions] = -1e20

            return np.argmax(q, axis=1)

    def update_target_net(self):
        self._target_net.load_state_dict(self._net.state_dict())
        self._target_net.eval()

    def _mini_batch_loop(self, buffer, grad_mngr):
        batch_pub_obs_t, \
        batch_a_t, \
        batch_range_idx, \
        batch_legal_action_mask_t, \
        batch_r_t, \
        batch_pub_obs_tp1, \
        batch_legal_action_mask_tp1, \
        batch_done = \
            buffer.sample(device=self.device, batch_size=self._args.batch_size)

        # [batch_size, n_actions]
        q1_t = self._net(pub_obses=batch_pub_obs_t, range_idxs=batch_range_idx,
                         legal_action_masks=batch_legal_action_mask_t.to(torch.float32))
        q1_tp1 = self._net(pub_obses=batch_pub_obs_tp1, range_idxs=batch_range_idx,
                           legal_action_masks=batch_legal_action_mask_tp1.to(torch.float32)).detach()
        q2_tp1 = self._target_net(pub_obses=batch_pub_obs_tp1, range_idxs=batch_range_idx,
                                  legal_action_masks=batch_legal_action_mask_tp1.to(torch.float32)).detach()

        # ______________________________________________ TD Learning _______________________________________________
        # [batch_size]
        q1_t_of_a_selected = q1_t[self._batch_arranged, batch_a_t]

        # only consider allowed actions for tp1
        q1_tp1 = torch.where(batch_legal_action_mask_tp1,
                             q1_tp1,
                             self._minus_e20)

        # [batch_size]
        _, best_a_tp1 = q1_tp1.max(dim=-1, keepdim=False)
        q2_best_a_tp1 = q2_tp1[self._batch_arranged, best_a_tp1]

        q2_best_a_tp1 = q2_best_a_tp1 * (1.0 - batch_done)
        target = batch_r_t + q2_best_a_tp1

        grad_mngr.backprop(pred=q1_t_of_a_selected, target=target)

    def state_dict(self):
        return {
            "q_net": self._net.state_dict(),
            "target_net": self._target_net.state_dict(),
            "eps": self._eps,
            "owner": self.owner,
            "args": self._args,
        }

    def load_state_dict(self, state):
        assert self.owner == state["owner"]
        # Not loading args by design

        self._net.load_state_dict(state["q_net"])
        self._target_net.load_state_dict(state["target_net"])
        self._eps = state["eps"]

    @staticmethod
    def from_state_dict(state_dict, env_bldr):
        ddqn = DDQN(owner=state_dict["owner"],
                    ddqn_args=state_dict["args"],
                    env_bldr=env_bldr)
        ddqn.load_state_dict(state_dict)
        ddqn.update_target_net()
        return ddqn

    @staticmethod
    def inference_version_from_state_dict(state_dict, env_bldr):
        ddqn = DDQN.from_state_dict(state_dict=state_dict, env_bldr=env_bldr)
        ddqn.buf = None
        ddqn.eps = None
        return ddqn


class DDQNArgs(_NetWrapperArgsBase):

    def __init__(self,
                 q_args,
                 cir_buf_size=1e5,
                 batch_size=512,
                 n_mini_batches_per_update=1,
                 target_net_update_freq=300,
                 optim_str="adam",
                 loss_str="mse",
                 lr=0.005,
                 eps_start=0.065,
                 eps_const=0.007,
                 eps_exponent=0.475,
                 eps_min=0.02,
                 grad_norm_clipping=10.0,
                 device_training="cpu",
                 ):
        assert isinstance(device_training, str), "Please pass a string (either 'cpu' or 'cuda')!"

        super().__init__(batch_size=batch_size,
                         n_mini_batches_per_update=n_mini_batches_per_update,
                         optim_str=optim_str,
                         loss_str=loss_str,
                         lr=lr,
                         grad_norm_clipping=grad_norm_clipping,
                         device_training=device_training)

        self.q_args = q_args
        self.cir_buf_size = int(cir_buf_size)
        self.target_net_update_freq = int(target_net_update_freq)
        self.eps_start = eps_start
        self.eps_const = eps_const
        self.eps_exponent = eps_exponent
        self.eps_min = eps_min
